"""
  Collection of codes for generating some training data sets.
"""

# Authors: B.J. Gross and P.J. Atzberger
# Website: http://atzberger.org/

import torch;
import numpy as np;
import pdb;

class diffOp1(torch.utils.data.Dataset):
  r"""
  Generates samples of the form :math:`(u^{[i]},f^{[i]})` where :math:`f^{[i]} = L[u^{[i]}]`, 
  where :math:`i` denotes the index of the sample.

  Stores data samples in the form :math:`(u,f)`.

  The samples of u are represented as a tensor of size [nsamples,nchannels,nx]
  and sample of f as a tensor of size [nsamples,nchannels,nx].
    
  Note:
    For now, please use nx that is odd.  In this initial implementation, we use a 
    method based on conjugated flips with formula for the odd case which is slightly
    simpler than other case.
    
  """
  def flipForFFT(self,u_k_part):
    r"""We flip as :math:`f_k = f_{N-k}`.  Notice that only :math:`0,\ldots,N-1` entries 
    stored.  This is useful for constructing real-valued function representations
    from random coefficients.  Real-valued function requires :math:`conj(f_k) = f_{N-k}`.
    We can use this flip to construct from random coefficients the term
    :math:`u_k = f_k + conj(flip(f_k))`,  then above constraint is satisfied.

    Args:
      a (Tensor): 1d array to flip.      
      
    Returns:
      Tensor: The flipped tensors symmetric under conjucation.      
    """
    nx = self.nx;        
    uu = u_k_part[:,:,nx:0:-1];
    vv = u_k_part[:,:,0];
    vv = np.expand_dims(vv,2);
    uu_k_flip = np.concatenate([vv,uu],2);     
    
    return uu_k_flip;

  def getComplex(self,a,b):
    j = np.complex(0,1); # create complex number (or use 1j).
    c = a + j*b;
    return c;

  def getRealImag(self,c):
    a = np.real(c);
    b = np.imag(c);
    return a,b;

  def computeLSymbol_ux(self):
    r"""Compute associated Fourier symbols for use under DFT for the operator L[u]."""
    nx = self.nx;
    vec_k1 = torch.zeros(nx);
    vec_k1_pp = torch.zeros(nx);    
    vec_k_sq = torch.zeros(nx);
    L_symbol_real = torch.zeros(nx,dtype=torch.float32);
    L_symbol_imag = torch.zeros(nx,dtype=torch.float32);
    two_pi = 2.0*np.pi;
    #two_pi_i = two_pi*1j; # $2\pi{i}$, 1j = sqrt(-1)
    for i in range(0,nx):
      vec_k1[i] = i;
      if (vec_k1[i] < nx/2):
        vec_k1_p = vec_k1[i];
      else:
        vec_k1_p = vec_k1[i] - nx;
      vec_k1_pp[i] = vec_k1_p;      
      L_symbol_real[i] = 0.0;
      L_symbol_imag[i] = two_pi*vec_k1_p;

      L_hat = self.getComplex(L_symbol_real.numpy(),L_symbol_imag.numpy());

    return L_hat, vec_k1_pp;

  def computeLSymbol_uxx(self):
    r"""Compute associated Fourier symbols for use under DFT for the operator L[u]."""
    nx = self.nx;
    vec_k1 = torch.zeros(nx);
    vec_k1_pp = torch.zeros(nx);    
    vec_k_sq = torch.zeros(nx);
    L_symbol_real = torch.zeros(nx,dtype=torch.float32);
    L_symbol_imag = torch.zeros(nx,dtype=torch.float32);
    neg_four_pi_sq = -4.0*np.pi*np.pi;    
    for i in range(0,nx):      
      vec_k1[i] = i;      
      vec_k_sq[i] = vec_k1[i]*vec_k1[i];
      if (vec_k1[i] < nx/2):
        vec_k1_p = vec_k1[i];
      else:
        vec_k1_p = vec_k1[i] - nx;
      vec_k1_pp[i] = vec_k1_p;
      vec_k_p_sq = vec_k1_p*vec_k1_p;
      L_symbol_real[i] = neg_four_pi_sq*vec_k_p_sq;
      L_symbol_imag[i] = 0.0;
        
      L_hat = self.getComplex(L_symbol_real.numpy(),L_symbol_imag.numpy());

    return L_hat, vec_k1_pp;

  def computeCoeffActionL(self,u_hat,L_hat):
    r"""Computes the action of operator L used for data generation in Fourier space."""
    u_k_real, u_k_imag = self.getRealImag(u_hat);
    L_symbol_real, L_symbol_imag = self.getRealImag(L_hat);

    f_k_real = L_symbol_real*u_k_real - L_symbol_imag*u_k_imag; #broadcast will distr over copies of u.    
    f_k_imag = L_symbol_real*u_k_imag + L_symbol_imag*u_k_real;

    # Generate samples u and f using ifft
    f_hat = self.getComplex(f_k_real,f_k_imag);

    return f_hat;

  def computeActionL(self,u,L_hat):
    r"""Computes the action of operator L used for data generation."""    
    raise Exception('Currently this routine not debugged, need to test first.')
    
    if flag_verbose > 0:
      print("computeActionL(): WARNING: Not yet fully tested.");
    
    # perform FFT to get u_hat
    u_hat = np.fft.fft(u);

    # compute action of L_hat
    f_hat = self.computeCoeffActionL(u_hat,L_hat);

    # compute inverse FFT to get f
    f = np.fft.ifft(f_hat);

    return f;

  def __init__(self,op_type='uxx',op_params=None,
               gen_mode='exp1',gen_params={'alpha1':0.1},
               num_samples=int(1e4),nchannels=1,nx=15,               
               flag_verbose=0, **extra_params):
    r"""Setup for data generation.
    
        Args:
          op_type (str): The differential operator to sample.
          op_params (dict): The operator parameters.
          gen_mode (str): The mode for the data generator.
          gen_params (dict): The parameters for the given generator.
          num_samples (int): The number of samples to generate.
          nchannels (int): The number of channels.
          nx (int): The number of input sample points. 
          flag_verbose (int): Level of reporting during calculations.
          extra_params (dict): Extra parameters for the sampler.
                                
        For extra_params we have:
          noise_factor (float): The amount of noise to add to samples.
          scale_factor (float): A factor to scale magnitude of the samples.
          flagComputeL (bool): If the fourier symbol of operator should be computed.
          
        For generator modes we have:
          gen_mode == 'exp1': 
            alpha1 (float): The decay rate.          

        Note:
          For now, please use only nx that is odd.  In this initial implementation, we use a 
          method based on conjugated flips with formula for the odd case which is slightly
          simpler than other case.
    """
    super(diffOp1, self).__init__();
    
    if flag_verbose > 0:
      print("Generating the data samples which can take some time.");
      print("num_samples = %d"%num_samples);

    self.op_type=op_type;
    self.op_params=op_params;

    self.gen_mode=gen_mode;
    self.gen_params=gen_params;
    
    self.num_samples=num_samples;
    self.nchannels=nchannels;
    self.nx=nx; 

    if (nx % 2 == 0):
      msg = "Not allowed yet to use nx that is even. ";
      msg += "For now, please just use nx that is odd given the flips currently used."
      raise Exception(msg);
    
    noise_factor=0;scale_factor=1.0;flagComputeL=False; # default values
    if 'noise_factor' in extra_params:
      noise_factor = extra_params['noise_factor'];

    if 'scale_factor' in extra_params:        
      scale_factor = extra_params['scale_factor'];

    if 'flagComputeL' in extra_params:        
      flagComputeL = extra_params['flagComputeL'];

    # Generate for the operator the Fourier symbols
    if self.op_type == 'ux' or self.op_type == 'u*ux' or self.op_type == 'ux*ux':
      L_hat, vec_k1_pp = self.computeLSymbol_ux();
    elif self.op_type == 'uxx' or self.op_type == 'u*uxx' or self.op_type == 'uxx*uxx':
      L_hat, vec_k1_pp = self.computeLSymbol_uxx();
    else:
      raise Exception("Unkonwn operator type.");

    if (flagComputeL):
      L_i        = np.fft.ifft(L_hat);
      self.L_hat = L_hat;
      self.L_i   = L_i;
      u          = np.zeros(nx);
      i0         = int(nx/2);      
      u[i0]      = 1.0;
      self.G_i   = self.computeActionL(u);

    # Generate random input function (want real-valued)
    # conj(u_k) = u_{N -k} needs to hold.
    u_k_real = np.random.randn(num_samples,nchannels,nx);
    u_k_imag = np.random.randn(num_samples,nchannels,nx);

    # scale modes to make smooth
    if gen_mode=='exp1':
      alpha1 = gen_params['alpha1'];
      factor_k = scale_factor*np.exp(-alpha1*vec_k1_pp**2);
      factor_k = factor_k.numpy();
    else:
      raise Exception("Generation mode not recognized.");
    
    u_k_real = u_k_real*factor_k; # broadcast will apply over last two dimensions
    u_k_imag = u_k_imag*factor_k; # broadcast will apply over last two dimensions    
    
    flag_debug = False;
    if flag_debug:
      if flag_verbose > 0:
        print("WARNING: debugging mode on.");
        
      u_k_real = 0.0*u_k_real;
      u_k_imag = 0.0*u_k_imag;
    
      u_k_real[0,0,1] = nx;
      u_k_imag[0,0,1] = 0;
    
      u_k_real[1,0,1] = 0;
      u_k_imag[1,0,1] = nx;
    
      u_k_real[2,0,1] = nx;
      u_k_imag[2,0,1] = nx;

    # flip modes for constructing rep of real-valued function    
    u_k_real_flip = self.flipForFFT(u_k_real);
    u_k_imag_flip = self.flipForFFT(u_k_imag);
    
    u_k_real_p = 0.5*u_k_real + 0.5*u_k_real_flip; # make conjugate conj(u_k) = u_{N -k}
    u_k_imag_p = 0.5*u_k_imag - 0.5*u_k_imag_flip; # make conjugate conj(u_k) = u_{N -k}
        
    u_k_real_p = torch.from_numpy(u_k_real_p);
    u_k_imag_p = torch.from_numpy(u_k_imag_p);
    
    u_k_real_p = u_k_real_p.type(torch.float32);
    u_k_imag_p = u_k_imag_p.type(torch.float32);
                    
    u_hat = self.getComplex(u_k_real_p.numpy(),u_k_imag_p.numpy());
        
    f_hat = self.computeCoeffActionL(u_hat,L_hat);
    f_hat = f_hat; # target operator relation for PDEs later is Lu = -f, so f = -Lu.
        
    # Generate samples u and f, in 2d using ifft2.    
    # ifft2 is broadcast over last two indices
    # perform inverse DFT to get u and f.
    u_i = np.fft.ifft(u_hat);
    f_i = np.fft.ifft(f_hat);
    
    if self.op_type == 'u*ux':
      f_i = u_i*f_i;
    elif self.op_type == 'ux*ux':
      f_i = f_i*f_i; 
    elif self.op_type == 'u*uxx':
      f_i = u_i*f_i;
    elif self.op_type == 'uxx*uxx':
      f_i = f_i*f_i;
        
    self.samples_X = torch.from_numpy(np.real(u_i)).type(torch.float32); # only grab real part
    self.samples_Y = torch.from_numpy(np.real(f_i)).type(torch.float32);
    
    if noise_factor > 0:
      self.samples_Y += noise_factor*torch.randn(*self.samples_Y.shape);
        
  def __len__(self):    
    return self.samples_X.size()[0];

  def __getitem__(self,index):
    return self.samples_X[index],self.samples_Y[index];

  def to(self,device):    
    self.samples_X = self.samples_X.to(device);
    self.samples_Y = self.samples_Y.to(device);
    
    return self;

class diffOp2(torch.utils.data.Dataset):
  r"""
  Generates samples of the form :math:`(u^{[i]},f^{[i]})` where :math:`f^{[i]} = L[u^{[i]}]`, 
  where :math:`i` denotes the index of the sample.

  Stores data samples in the form :math:`(u,f)`.

  The samples of u are represented as a tensor of size [nsamples,nchannels,nx]
  and sample of f as a tensor of size [nsamples,nchannels,nx].
   
  Note:
    For now, please use nx that is odd.  In this initial implementation, we use a 
    method based on conjugated flips with formula for the odd case which is slightly
    simpler than other case.
    
  """
  def flipForFFT(self,u_k_part):
    r"""We flip as :math:`f_k = f_{N-k}`.  Notice that only :math:`0,\ldots,N-1` entries 
    stored.  This is useful for constructing real-valued function representations
    from random coefficients.  Real-valued function requires :math:`conj(f_k) = f_{N-k}`.
    We can use this flip to construct from random coefficients the term
    :math:`u_k = f_k + conj(flip(f_k))`,  then above constraint is satisfied.

    Args:
      a (Tensor): 1d array to flip.      
      
    Returns:
      Tensor: The flipped tensors symmetric under conjucation.      
    """
    nx = self.nx;ny = self.ny;
    
    u_k_part_row0 = u_k_part[:,:,0,:];
    u_k_part_row0 = np.expand_dims(u_k_part_row0,2);
    u_k_part_ex = np.concatenate([u_k_part,u_k_part_row0],2);

    u_k_part_col0 = u_k_part_ex[:,:,:,0];
    u_k_part_col0 = np.expand_dims(u_k_part_col0,3);
    u_k_part_ex = np.concatenate([u_k_part_ex,u_k_part_col0],3);

    u_k_part_ex_flip = np.flip(u_k_part_ex,2);
    u_k_part_ex_flip = np.flip(u_k_part_ex_flip,3);

    u_k_part_flip = np.delete(u_k_part_ex_flip,nx,2);
    u_k_part_flip = np.delete(u_k_part_flip,ny,3);

    return u_k_part_flip;

  def getComplex(self,a,b):
    j = np.complex(0,1); # create complex number (or use 1j).
    c = a + j*b;
    return c;

  def getRealImag(self,c):
    a = np.real(c);
    b = np.imag(c);
    return a,b;

  def computeLSymbol_laplacian_u(self):
    r"""Compute associated Fourier symbols for use under DFT for the operator L[u]."""
    num_dim = 1;nx=self.nx;ny=self.ny;
    vec_k1 = torch.zeros(nx,ny);
    vec_k2 = torch.zeros(nx,ny);    
    vec_k1_pp = torch.zeros(nx,ny);
    vec_k2_pp = torch.zeros(nx,ny);
    vec_k_sq = torch.zeros(nx,ny);
    L_symbol_real = torch.zeros(nx,ny,dtype=torch.float32);
    L_symbol_imag = torch.zeros(nx,ny,dtype=torch.float32);
    neg_four_pi_sq = -4.0*np.pi*np.pi;
    for i in range(0,nx):
      for j in range(0,ny):
        vec_k1[i,j] = i;
        vec_k2[i,j] = j;
        vec_k_sq[i,j] = vec_k1[i,j]*vec_k1[i,j] + vec_k2[i,j]*vec_k2[i,j];
        if (vec_k1[i,j] < nx/2):
          vec_k1_p = vec_k1[i,j];
        else:
          vec_k1_p = vec_k1[i,j] - nx;  
        if (vec_k2[i,j] < ny/2):
          vec_k2_p = vec_k2[i,j];
        else:
          vec_k2_p = vec_k2[i,j] - ny; 
        vec_k1_pp[i,j] = vec_k1_p;
        vec_k2_pp[i,j] = vec_k2_p;
        vec_k_p_sq = vec_k1_p*vec_k1_p + vec_k2_p*vec_k2_p;
        L_symbol_real[i,j] = neg_four_pi_sq*vec_k_p_sq;
        L_symbol_imag[i,j] = 0.0;
        
    L_hat = self.getComplex(L_symbol_real.numpy(),L_symbol_imag.numpy());
        
    return L_hat, vec_k1_pp, vec_k2_pp;

  def computeLSymbol_grad_u(self):
    r"""Compute associated Fourier symbols for use under DFT for the operator L[u]."""
    num_dim = 2;nx=self.nx;ny=self.ny;
    vec_k1 = torch.zeros(nx,ny);
    vec_k2 = torch.zeros(nx,ny);    
    vec_k1_pp = torch.zeros(nx,ny);
    vec_k2_pp = torch.zeros(nx,ny);
    vec_k_sq = torch.zeros(nx,ny);
    L_symbol_real = torch.zeros(num_dim,nx,ny,dtype=torch.float32);
    L_symbol_imag = torch.zeros(num_dim,nx,ny,dtype=torch.float32);
    two_pi = 2.0*np.pi;
    #two_pi_i = two_pi*1j; # $2\pi{i}$, 1j = sqrt(-1)
    for i in range(0,nx):
      for j in range(0,ny):
        vec_k1[i,j] = i;
        vec_k2[i,j] = j;
        vec_k_sq[i,j] = vec_k1[i,j]*vec_k1[i,j] + vec_k2[i,j]*vec_k2[i,j];
        if (vec_k1[i,j] < nx/2):
          vec_k1_p = vec_k1[i,j];
        else:
          vec_k1_p = vec_k1[i,j] - nx;  
        if (vec_k2[i,j] < ny/2):
          vec_k2_p = vec_k2[i,j];
        else:
          vec_k2_p = vec_k2[i,j] - ny; 
        vec_k1_pp[i,j] = vec_k1_p;
        vec_k2_pp[i,j] = vec_k2_p;
        vec_k_p_sq = vec_k1_p*vec_k1_p + vec_k2_p*vec_k2_p;
        L_symbol_real[0,i,j] = 0.0;
        L_symbol_imag[0,i,j] = two_pi*vec_k1_p;
        L_symbol_real[1,i,j] = 0.0;
        L_symbol_imag[1,i,j] = two_pi*vec_k2_p;
        
    L_hat_0 = self.getComplex(L_symbol_real[0,:,:].numpy(),L_symbol_imag[0,:,:].numpy());
    L_hat_1 = self.getComplex(L_symbol_real[1,:,:].numpy(),L_symbol_imag[1,:,:].numpy());

    L_hat   = np.stack((L_hat_0,L_hat_1));
       
    return L_hat, vec_k1_pp, vec_k2_pp;

  def computeCoeffActionL(self,u_hat,L_hat):
    r"""Computes the action of operator L used for data generation in Fourier space."""
    u_k_real, u_k_imag = self.getRealImag(u_hat);    
    L_symbol_real, L_symbol_imag = self.getRealImag(L_hat);
    
    f_k_real = L_symbol_real*u_k_real - L_symbol_imag*u_k_imag; #broadcast will distr over copies of u.
    #f_k_real = -1.0*f_k_real;
    f_k_imag = L_symbol_real*u_k_imag + L_symbol_imag*u_k_real;
    #f_k_imag = -1.0*f_k_imag;
    
    # Generate samples u and f using ifft2.    
    f_hat = self.getComplex(f_k_real,f_k_imag);
        
    return f_hat;

  def computeActionL(self,u,L_hat):
    r"""Computes the action of operator L used for data generation."""    
    raise Exception('Currently this routine not debugged, need to test first.')
    
    # perform FFT to get u_hat
    u_hat = np.fft.fft2(u);

    # compute action of L_hat
    f_hat = self.computeCoeffActionL(u_hat,L_hat);

    # compute inverse FFT to get f
    f = np.fft.ifft2(f_hat)

    return f;
                
  def __init__(self,op_type=r'\Delta{u}',op_params=None,
               gen_mode='exp1',gen_params={'alpha1':0.1},
               num_samples=int(1e4),nchannels=1,nx=15,ny=15,
               flag_verbose=0, **extra_params):
    r"""Setup for data generation.
    
        Args:
          op_type (str): The differential operator to sample.
          op_params (dict): The operator parameters.
          gen_mode (str): The mode for the data generator.
          gen_params (dict): The parameters for the given generator.
          num_samples (int): The number of samples to generate.
          nchannels (int): The number of channels.
          nx (int): The number of input sample points in x-direction.
          ny (int): The number of input sample points in y- direction.
          flag_verbose (int): Level of reporting during calculations.
          extra_params (dict): Extra parameters for the sampler.
                                
        For extra_params we have:
          noise_factor (float): The amount of noise to add to samples.
          scale_factor (float): A factor to scale magnitude of the samples.
          flagComputeL (bool): If the fourier symbol of operator should be computed.
          
        For generator modes we have:
          gen_mode == 'exp1': 
            alpha1 (float): The decay rate.          

        Note:
          For now, please use only nx that is odd.  In this initial implementation, we use a 
          method based on conjugated flips with formula for the odd case which is slightly
          simpler than other case.
    """
    if flag_verbose > 0:
      print("Generating the data samples which can take some time.");
      print("num_samples = %d"%num_samples);

    self.op_type=op_type;
    self.op_params=op_params;

    self.gen_mode=gen_mode;
    self.gen_params=gen_params;
    
    self.num_samples=num_samples;
    self.nchannels=nchannels;
    self.nx=nx; self.ny=ny;

    if (nx % 2 == 0) or (ny % 2 == 0) or (nx != ny): # may be able to relax nx != ny (just for safety)
      msg = "Not allowed yet to use nx,ny that are even or unequal. ";
      msg += "For now, please just use nx,ny that is odd given the flips currently used."
      raise Exception(msg);
    
    noise_factor=0;scale_factor=1.0;flagComputeL=False; # default values
    if 'noise_factor' in extra_params:
      noise_factor = extra_params['noise_factor'];

    if 'scale_factor' in extra_params:        
      scale_factor = extra_params['scale_factor'];

    if 'flagComputeL' in extra_params:        
      flagComputeL = extra_params['flagComputeL'];

    # Generate for the operator the Fourier symbols
    flag_vv = 'null';
    if self.op_type == r'\grad{u}' or self.op_type == r'u\grad{u}' or self.op_type == r'\grad{u}\cdot\grad{u}':
      L_hat, vec_k1_pp, vec_k2_pp = self.computeLSymbol_grad_u();
      flag_vv = 'vector2';
    elif self.op_type == r'\Delta{u}' or self.op_type == r'u\Delta{u}' or self.op_type == r'\Delta{u}*\Delta{u}':
      L_hat, vec_k1_pp, vec_k2_pp = self.computeLSymbol_laplacian_u();
      flag_vv = 'scalar';
    else:
      raise Exception("Unknown operator type.");
             
    if (flagComputeL):
      raise Exception("Currently not yet supported, the flagComputeL.");
      L_i = np.fft.ifft2(L_hat);
      self.L_hat = L_hat;
      self.L_i   = L_i;
      u          = np.zeros(nx,ny);
      i0         = int(nx/2);
      j0         = int(ny/2);
      u[i0,j0]   = 1.0;
      self.G_i   = self.computeActionL(u);

    # Generate random input function (want real-valued)
    # conj(u_k) = u_{N -k} needs to hold.
    u_k_real = np.random.randn(num_samples,nchannels,nx,ny);
    u_k_imag = np.random.randn(num_samples,nchannels,nx,ny);
    
    # scale modes to make smooth
    if gen_mode=='exp1':
      alpha1 = gen_params['alpha1'];
      factor_k = scale_factor*np.exp(-alpha1*(vec_k1_pp**2 + vec_k2_pp**2));
      factor_k = factor_k.numpy();
    else:
      raise Exception("Generation mode not recognized.");
       
    u_k_real = u_k_real*factor_k; # broadcast will apply over last two dimensions
    u_k_imag = u_k_imag*factor_k; # broadcast will apply over last two dimensions

    # flip modes for constructing rep of real-valued function
    u_k_real_flip = self.flipForFFT(u_k_real);
    u_k_imag_flip = self.flipForFFT(u_k_imag);
    
    u_k_real = 0.5*u_k_real + 0.5*u_k_real_flip; # make conjugate conj(u_k) = u_{N -k}
    u_k_imag = 0.5*u_k_imag - 0.5*u_k_imag_flip; # make conjugate conj(u_k) = u_{N -k}
        
    u_k_real = torch.from_numpy(u_k_real);
    u_k_imag = torch.from_numpy(u_k_imag);
    
    u_k_real = u_k_real.type(torch.float32);
    u_k_imag = u_k_imag.type(torch.float32);
                    
    u_hat = self.getComplex(u_k_real.numpy(),u_k_imag.numpy());
    if flag_vv == 'scalar':
      f_hat = self.computeCoeffActionL(u_hat,L_hat);
    elif flag_vv == 'vector2':
      f_hat_0 = self.computeCoeffActionL(u_hat,L_hat[0,:,:]);
      f_hat_1 = self.computeCoeffActionL(u_hat,L_hat[1,:,:]);
      f_hat   = np.concatenate((f_hat_0,f_hat_1),-3);
    else:
      raise Exception("Unkonwn operator type.");

    # Generate samples u and f using ifft2.
    # ifft2 is broadcast over last two indices
    # perform inverse DFT to get u and f
    u_i = np.fft.ifft2(u_hat);
    if flag_vv == 'scalar':
      f_i = np.fft.ifft2(f_hat);
    elif flag_vv == 'vector2':
      f_i_0 = np.fft.ifft2(f_hat[:,0,:,:]);
      f_i_1 = np.fft.ifft2(f_hat[:,1,:,:]);
      f_i   = np.stack((f_i_0,f_i_1),-3);
    else:
      raise Exception("Unkonwn operator type.");

    if self.op_type == r'\grad{u}':
      f_i = f_i; # nothing to do.
    elif self.op_type == r'u\grad{u}':
      f_i = u_i*f_i; # matches up by broadcast rules
    elif self.op_type == r'\grad{u}\cdot\grad{u}':
      f_i = np.sum(f_i**2,1); # sum on axis for channels, [batch,channel,nx,ny].
      f_i = np.expand_dims(f_i,1); # keep in form [batch,1,nx,ny]
    elif self.op_type == r'\Delta{u}':
      f_i = f_i;  # nothing to do.
    elif self.op_type == r'u\Delta{u}':
      f_i = u_i*f_i; 
    elif self.op_type == r'\Delta{u}*\Delta{u}':
      f_i = f_i**2; 
    else:
      raise Exception("Unkonwn operator type.");

    self.samples_X = torch.from_numpy(np.real(u_i)).type(torch.float32); # only grab real part
    self.samples_Y = torch.from_numpy(np.real(f_i)).type(torch.float32);

    if noise_factor > 0:
      self.samples_Y += noise_factor*torch.randn(*self.samples_Y.shape);

  def __len__(self):
    return self.samples_X.size()[0];

  def __getitem__(self,index):
    return self.samples_X[index],self.samples_Y[index];

  def to(self,device):
    self.samples_X = self.samples_X.to(device);
    self.samples_Y = self.samples_Y.to(device);
    return self;




"""
  .. image:: overview.png

  PyTorch implementation of GMLS-Nets.  Module for neural networks for 
  processing scattered data sets using Generalized Moving Least Squares (GMLS).
 
  If you find these codes or methods helpful for your project, please cite:

  |   @article{trask_patel_gross_atzberger_GMLS_Nets_2019,
  |     title={GMLS-Nets: A framework for learning from unstructured data},
  |     author={Nathaniel Trask, Ravi G. Patel, Ben J. Gross, Paul J. Atzberger},
  |     journal={arXiv:1909.05371},  
  |     month={September},
  |     year={2019},
  |     url={https://arxiv.org/abs/1909.05371}
  |   }

"""

# Authors: B.J. Gross and P.J. Atzberger
# Website: http://atzberger.org/

import torch;
import torch.nn as nn;
import torchvision;
import torchvision.transforms as transforms;

import numpy as np;

import scipy.spatial as spatial # used for finding neighbors within distance $\delta$

from collections import OrderedDict;

import pickle as p;

import pdb;

import time;

# ====================================
# Custom Functions
# ====================================
class MapToPoly_Function(torch.autograd.Function):
  r"""
  This layer processes a collection of scattered data points consisting of a collection 
  of values :math:`u_j` at points :math:`x_j`.  For a collection of target points 
  :math:`x_i`, local least-squares problems are solved for obtaining a local representation 
  of the data over a polynomial space.  The layer outputs a collection of polynomial 
  coefficients :math:`c(x_i)` at each point and the collection of target points :math:`x_i`.
  """
  @staticmethod
  def weight_one_minus_r(z1,z2,params):  
    r"""Weight function :math:`\omega(x_j,x_i) = \left(1 - r/\epsilon\right)^{\bar{p}}_+.` 

        Args:
          z1 (Tensor): The first point.  Tensor of shape [1,num_dims].
          z2 (Tensor): The second point.    Tensor of shape [1,num_dims].
          params (dict): The parameters are 'p' for decay power and 'epsilon' for support size.

        Returns:
          Tensor: The weight evaluation over points.
     """
    epsilon = params['epsilon']; p = params['p'];
    r = torch.sqrt(torch.sum(torch.pow(z1 - z2,2),1));          
      
    diff    = torch.clamp(1 - (r/epsilon),min=0);
    eval    = torch.pow(diff,p);
    return eval;

  @staticmethod
  def get_num_polys(porder,num_dim=None):
    r""" Returns the number of polynomials of given porder. """
    if num_dim == 1:
      num_polys = porder + 1;
    elif num_dim == 2:
      num_polys = int((porder + 2)*(porder + 1)/2);
    elif num_dim == 3:
      num_polys = 0;
      for beta in range(0,porder + 1):
        num_polys += int((porder - beta + 2)*(porder - beta + 1)/2);
    else:
      raise Exception("Number of dimensions not implemented currently. \n  num_dim = %d."%num_dim);
  
    return num_polys;

  @staticmethod
  def eval_poly(pts_x,pts_x2_i0,c_star_i0,porder,flag_verbose):
    r""" Evaluates the polynomials locally around a target point xi given coefficients c. """
    # Evaluates the polynomial locally (this helps to assess the current fit).
    # Implemented for 1D, 2D, and 3D.  
    #
    # 2D:
    # Computes Taylor Polynomials over x and y.
    # T_{k1,k2}(x1,x2) = (1.0/(k1 + k2)!)*(x1 - x01)^{k1}*(x2 - x02)^{k2}.        
    # of terms is N = (porder + 1)*(porder + 2)/2.
    #
    # WARNING: Note the role of factorials and orthogonality here.  The Taylor 
    # expansion/polynomial formulation is not ideal and can give ill-conditioning.  
    # It would be better to use orthogonal polynomials or other bases.
    #
    num_dim = pts_x.shape[1];
    if num_dim == 1:
      II = 0;
      alpha_factorial = 1.0;
      eval_p = torch.zeros(pts_x.shape[0],device=c_star_i0.device);
      for alpha in np.arange(0,porder + 1):
        if alpha >= 2:
          alpha_factorial *= alpha;
        if flag_verbose > 1: print("alpha = " + str(alpha)); print("k = " + str(k));
        # for now, (x - x_*)^k, but ideally use orthogonal polynomials
        base_poly = torch.pow(pts_x[:,0] - pts_x2_i0[0],alpha);
        base_poly = base_poly/alpha_factorial;
        eval_p   += c_star_i0[II]*base_poly;
        II       += 1;
    elif num_dim == 2:
      II = 0;
      alpha_factorial = 1.0;
      eval_p = torch.zeros(pts_x.shape[0],device=c_star_i0.device);
      for alpha in np.arange(0,porder + 1):
        if alpha >= 2:
          alpha_factorial *= alpha;
        for k in np.arange(0,alpha + 1):
          if flag_verbose > 1: print("alpha = " + str(alpha)); print("k = " + str(k));
          # for now, (x - x_*)^k, but ideally use orthogonal polynomials
          base_poly = torch.pow(pts_x[:,0] - pts_x2_i0[0],alpha - k);
          # for now, (x - x_*)^k, but ideally use orthogonal polynomials
          base_poly = base_poly*torch.pow(pts_x[:,1] - pts_x2_i0[1],k);
          base_poly = base_poly/alpha_factorial;
          eval_p   += c_star_i0[II]*base_poly;
          II       += 1;
    elif num_dim == 3: # caution, below gives initial results, but should be more fully validated
      II = 0;
      alpha_factorial = 1.0;
      eval_p = torch.zeros(pts_x.shape[0],device=c_star_i0.device);
      for beta in np.arange(0,porder + 1):
        base_poly = torch.pow(pts_x[:,2] - pts_x2_i0[2],beta);
        for alpha in np.arange(0,porder - beta + 1):
          if alpha >= 2:
            alpha_factorial *= alpha;
          for k in np.arange(0,alpha + 1):
            if flag_verbose > 1: print("alpha = " + str(alpha)); print("k = " + str(k));
            # for now, (x - x_*)^k, but ideally use orthogonal polynomials
            base_poly = base_poly*torch.pow(pts_x[:,0] - pts_x2_i0[0],alpha - k);
            base_poly = base_poly*torch.pow(pts_x[:,1] - pts_x2_i0[1],k);
            base_poly = base_poly/alpha_factorial;
            eval_p   += c_star_i0[II]*base_poly;
            II       += 1;
    else:
      raise Exception("Number of dimensions not implemented currently. \n  num_dim = %d."%num_dim);

    return eval_p;

  @staticmethod
  def generate_mapping(weight_func,weight_func_params,
                       porder,epsilon,
                       pts_x1,pts_x2,
                       tree_points=None,device=None,
                       flag_verbose=0):
    r""" Generates for caching the data for the mapping from field values (uj,xj) :math:`\rightarrow` (ci,xi).  
    This help optimize codes and speed up later calculations that are done repeatedly."""
    if device is None:
      device = torch.device('cpu');
      
    map_data = {};
  
    num_dim = pts_x1.shape[1];
  
    if pts_x2 is None:
      pts_x2 = pts_x1;
	
    pts_x1 = pts_x1.to(device);
    pts_x2 = pts_x2.to(device);      

    pts_x1_numpy = None; pts_x2_numpy = None; 
    if tree_points is None: # build kd-tree of points for neighbor listing
      if pts_x1_numpy is None: pts_x1_numpy = pts_x1.cpu().numpy();
      tree_points = spatial.cKDTree(pts_x1_numpy);  

    # Maps from u(x_j) on $x_j \in \mathcal{S}^1$ to a 
    # polynomial representations in overlapping regions $\Omega_i$ at locations
    # around points $x_i \in \mathcal{S}^2$.
    # These two sample sets need not be the same allowing mappings between point locations.
    # Computes polynomials over x and y.
    # Number of terms in 2D is num_polys = (porder + 1)*(porder + 2)/2.
    num_pts1 = pts_x1.shape[0]; num_pts2 = pts_x2.shape[0];
    num_polys = MapToPoly_Function.get_num_polys(porder,num_dim);
    if flag_verbose > 0: 
      print("num_polys = " + str(num_polys));
      
    M = torch.zeros((num_pts2,num_polys,num_polys),device=device); # assemble matrix at each grid-point
    M_inv = torch.zeros((num_pts2,num_polys,num_polys),device=device); # assemble matrix at each grid-point
    
    #svd_U = torch.zeros((num_pts2,num_polys,num_polys)); # assemble matrix at each grid-point
    #svd_S = torch.zeros((num_pts2,num_polys,num_polys)); # assemble matrix at each grid-point
    #svd_V = torch.zeros((num_pts2,num_polys,num_polys)); # assemble matrix at each grid-point
    
    vec_rij = torch.zeros((num_pts2,num_polys,num_pts1),device=device); # @optimize: ideally should be sparse matrix.

    # build up the batch of linear systems for each target point
    for i in np.arange(0,num_pts2): # loop over the points $x_i$
        
      if (flag_verbose > 0) & (i % 100 == 0): print("i = " + str(i) + " : num_pts2 = " + str(num_pts2));

      if pts_x2_numpy is None: pts_x2_numpy = pts_x2.cpu().numpy();
      indices_xj_i = tree_points.query_ball_point(pts_x2_numpy[i,:], epsilon); # find all points with distance 
                                                                               # less than epsilon from xi.

      for j in indices_xj_i: # @optimize later to use only local points, and where weights are non-zero.
          
        if flag_verbose > 1: print("j = " + str(j));
        
        vec_p_j = torch.zeros(num_polys,device=device);
        w_ij    = weight_func(pts_x1[j,:].unsqueeze(0), pts_x2[i,:].unsqueeze(0), weight_func_params);  # can optimize for sub-lists outer-product          

        # Computes Taylor Polynomials over x,y,z.
        #
				# 2D Case:
        #   T_{k1,k2}(x1,x2) = (1.0/(k1 + k2)!)*(x1 - x01)^{k1}*(x2 - x02)^{k2}.
        #   number of terms is N = (porder + 1)*(porder + 2)/2.
        #   computes polynomials over x and y.
        #
        # WARNING: The monomial basis is non-ideal and can lead to ill-conditioned linear algebra.
        # This ultimately should be generalized in the future to other bases, ideally orthogonal, 
        # which would help both with efficiency and conditioning of the linear algebra.
        #
        if num_dim == 1:
          # number of terms is N = porder + 1.
          II = 0;
          for alpha in np.arange(0,porder + 1):
            if flag_verbose > 1: print("alpha = " + str(alpha)); print("k = " + str(k));
            # for now, (x - x_*)^k, but ideally use orthogonal polynomials
            vec_p_j[II] = torch.pow(pts_x1[j,0] - pts_x2[i,0], alpha);
            II += 1;
        elif num_dim == 2:
          # number of terms is N = (porder + 1)*(porder + 2)/2.
          II = 0;
          for alpha in np.arange(0,porder + 1):
            for k in np.arange(0,alpha + 1):
              if flag_verbose > 1: print("alpha = " + str(alpha)); print("k = " + str(k));
              # for now, (x - x_*)^k, but ideally use orthogonal polynomials
              vec_p_j[II] = torch.pow(pts_x1[j,0] - pts_x2[i,0], alpha - k);
              vec_p_j[II] = vec_p_j[II]*torch.pow(pts_x1[j,1] - pts_x2[i,1], k);
              II += 1;
        elif num_dim == 3:
          # number of terms is N = sum_{alpha_3 = 0}^porder [(porder - alpha_3+ 1)*(porder - alpha_3 + 2)/2.
          II = 0;
          for beta in np.arange(0,porder + 1):
            vec_p_j[II] = torch.pow(pts_x1[j,2] - pts_x2[i,2],beta);
            for alpha in np.arange(0,porder - beta + 1):
              for k in np.arange(0,alpha + 1):
                if flag_verbose > 1:
                  print("beta = " + str(beta)); print("alpha = " + str(alpha)); print("k = " + str(k));
                # for now, (x - x_*)^k, but ideally use orthogonal polynomials
                vec_p_j[II] = vec_p_j[II]*torch.pow(pts_x1[j,0] - pts_x2[i,0],alpha - k);
                vec_p_j[II] = vec_p_j[II]*torch.pow(pts_x1[j,1] - pts_x2[i,1],k);
                II += 1;
                          
        # add contributions to the M(x_i) and r(x_i) terms
        # r += (w_ij*u[j])*vec_p_j;
        vec_rij[i,:,j] = w_ij*vec_p_j;
        M[i,:,:]      += torch.ger(vec_p_j,vec_p_j)*w_ij;  # outer-product of vectors (build match of matrices)

      # Compute the SVD of M for purposes of computing the pseudo-inverse (for solving least-squares problem).
      # Note: M is always symmetric positive semi-definite, so U and V should be transposes of each other
      # and sigma^2 are the eigenvalues squared.  This simplifies some expressions.
      
      U,S,V = torch.svd(M[i,:,:]); # M = U*SS*V^T, note SS = diag(S)
      threshold_nonzero = 1e-9; # threshold for the largest singular value to consider being non-zero.
      I_nonzero         = (S > threshold_nonzero);
      S_inv             = 0.0*S;
      S_inv[I_nonzero]  = 1.0/S[I_nonzero];
      SS_inv            = torch.diag(S_inv);
      M_inv[i,:,:]      = torch.matmul(V,torch.matmul(SS_inv,U.t())); # pseudo-inverse of M^{-1} = V*S^{-1}*U^T

    # Save the linear system information for the least-squares problem at each target point $xi$.
    map_data['M'] = M;
    map_data['M_inv'] = M_inv;
    map_data['vec_rij'] = vec_rij;
    
    return map_data;   

  @staticmethod
  def get_poly_1D_u(u, porder, weight_func, weight_func_params,
                    pts_x1, epsilon = None, pts_x2 = None, cached_data=None,
                    tree_points = None, device=None, flag_verbose = 0):
      r""" Compute the polynomial coefficients in the case of a scalar field.  Would not typically call directly, used for internal purposes. """

      # We assume that all inputs are pytorch tensors
      # Assumes:
      #  pts_x1.size = [num_pts,num_dim] 
      #  pts_x2.size = [num_pts,num_dim]
      # 
      # @optimize: Should cache the points and neighbor lists... then using torch.solve, torch.ger.
      #            Should vectorize all of the for-loop operations via Lambdifying polynomial evals.
      #            Should avoid numpy calculations, maybe cache numpy copy of data if needed to avoid .cpu() transfer calls.
      #            Use batching over points to do solves, then GPU parallizable and faster.
      #      
      if device is None:
        device = torch.device('cpu'); # default cpu device
  
      if (u.dim() > 1):
        print("u.dim = " + str(u.dim()));
        print("u.shape = " + str(u.shape));
        raise Exception("Assumes input with dimension == 1.");
      
      if (cached_data is None) or ('map_data' not in cached_data) or (cached_data['map_data'] is None):
        generate_mapping = MapToPoly_Function.generate_mapping;
        
        if pts_x2 is None:
          pts_x2 = pts_x1;
          
        map_data = generate_mapping(weight_func,weight_func_params,
                                    porder,epsilon,
                                    pts_x1,pts_x2,tree_points,device);
        
        if cached_data is not None:
          cached_data['map_data'] = map_data;
          
      else:
        map_data = cached_data['map_data']; # use cached data
      
      if flag_verbose > 0: 
        print("num_pts1 = " + str(num_pts1) + ", num_pts2 = " + str(num_pts2));

      if epsilon is None:
        raise Exception('The epsilon ball size to use around xi must be specified.')
                
      # Maps from u(x_j) on $x_j \in \mathcal{S}^1$ to a 
      # polynomial representations in overlapping regions $\Omega_i$ at locations
      # around points $x_i \in \mathcal{S}^2$.
      
      # These two sample sets need not be the same allowing mappings between point sets.
      # Computes polynomials over x and y.
      # For 2D case, number of terms is num_polys = (porder + 1)*(porder + 2)/2.

      #c_star[:,i] = np.linalg.solve(np_M,np_r); # "c^*(x_i) = M^{-1}*r."
      vec_rij = map_data['vec_rij'];
      M_inv   = map_data['M_inv'];

      r_all   = torch.matmul(vec_rij,u);
      c_star  = torch.bmm(M_inv,r_all.unsqueeze(2)); # perform batch matric-vector multiplications
      c_star  = c_star.squeeze(2); # convert to list of vectors
      
      output  = c_star;
      output  = output.float(); # Map to float type for GPU / PyTorch Module compatibilities.
      
      return output, pts_x2;

  @staticmethod
  def forward(ctx, input, porder, weight_func, weight_func_params,
              pts_x1, epsilon = None, pts_x2 = None, cached_data=None,
              tree_points = None, device = None, flag_verbose = 0):
      r"""

      For a field u specified at points xj, performs the mapping to coefficients c at points xi, (uj,xj) :math:`\rightarrow` (ci,xi).

      Args:
        input (Tensor): The input field data uj.
        porder (int): Order of the basis to use (polynomial degree).
        weight_func (function): Weight function to use.
        weight_func_params (dict): Weight function parameters.
        pts_x1 (Tensor): The collection of domain points :math:`x_j`.
        epsilon (float): The :math:`\epsilon`-neighborhood size to use to sort points (should be compatible with choice of weight_func_params).
        pts_x2 (Tensor): The collection of target points :math:`x_i`.
        cache_data (dict): Stored data to help speed up repeated calculations.
        tree_points (dict): Stored data to help speed up repeated calculations.
        device (torch.device): Device on which to perform calculations (GPU or other, default is CPU).
        flag_verbose (int): Level of reporting on progress during the calculations.
        
      Returns: 
        tuple of (ci,xi): The coefficient values ci at the target points xi.  The target points xi. 
      
      """
      if device is None:
        device = torch.device('cpu');

      ctx.atz_name = 'MapToPoly_Function';

      ctx.save_for_backward(input,pts_x1,pts_x2);

      ctx.atz_porder = porder;
      ctx.atz_weight_func = weight_func;
      ctx.atz_weight_func_params = weight_func_params;

      get_poly_1D_u = MapToPoly_Function.get_poly_1D_u;
      get_num_polys = MapToPoly_Function.get_num_polys;
      
      input_dim = input.dim();
      if input_dim >= 1: # compute c_star in batches
          pts_x1_numpy = None;
          pts_x2_numpy = None; 
          
          if pts_x2 is None:
            pts_x2 = pts_x1;
          
          # reshape the data to handle as a batch [batch_size, uj_data_size]
          # We assume u is input in the form [I,k,xj], u_I(k,xj), the index I is arbitrary.
          u         = input;
          
          if input_dim == 2: # need to unsqueeze, so 2D we are mapping
                             # [k,xj] --> [I,k,xj] --> [II,c] --> [I,k,xi,c] --> [k,xi,c] 
            u = u.unsqueeze(0); # u(k,xj) assumed in our calculations here
          
          if input_dim == 1: # need to unsqueeze, so 1D we are mapping
                             # [xj] --> [I,k,xj] --> [II,c] --> [I,k,xi,c] --> [xi,c] 
            u = u.unsqueeze(0); # u(k,xj) assumed in our calculations here
            u = u.unsqueeze(0); # u(k,xj) assumed in our calculations here      
            
          u_num_dim = u.dim();            
          size_nm1  = 1;
          for d in range(u_num_dim - 1): 
            size_nm1 *= u.shape[d];

          uu = u.contiguous().view((size_nm1,u.shape[-1]));

          # compute the sizes of c_star and number of points
          num_dim   = pts_x1.shape[1];
          num_polys = get_num_polys(porder,num_dim);
          num_pts2  = pts_x2.shape[0];

          # output needs to be of size [batch_size, xi_data_size, num_polys]
          output = torch.zeros((uu.shape[0],num_pts2,num_polys),device=device); # will reshape at the end

          # loop over the batches and compute the c_star in each case
          if cached_data is None:
            cached_data = {}; # create empty, which can be computed first time to store data.
            
          if tree_points is None:
            if pts_x1_numpy is None: pts_x1_numpy = pts_x1.cpu().numpy();
            tree_points = spatial.cKDTree(pts_x1_numpy);
                              
          for k in range(uu.shape[0]):
            uuu = uu[k,:];
            out, pts_x2 = get_poly_1D_u(uuu,porder,weight_func,weight_func_params,
                                        pts_x1,epsilon,pts_x2,cached_data,
                                        tree_points,flag_verbose);
            output[k,:,:] = out;

          # final output should be [*, xi_data_size, num_polys], where * is the original sizes
          # for indices [i1,i2,...in,k_channel,xi_data,c_poly_coeff].
          output = output.view(*u.shape[0:u_num_dim-1],num_pts2,num_polys);

          if input_dim == 2: # 2D special case we just return k, xi, c (otherwise feed input 3D [I,k,u(xj)] I=1,k=1).
            output = output.squeeze(0);
          
          if input_dim == 1: # 1D special case we just return xi, c (otherwise feed input 3D [I,k,u(xj)] I=1,k=1).
            output = output.squeeze(0);
            output = output.squeeze(0);          
                          
      else:
        print("input.dim = " + str(input.dim()));
        print("input.shape = " + str(input.shape));
        raise Exception("input tensor dimension not yet supported, only dim = 1 and dim = 3 currently.");

      ctx.atz_cached_data = cached_data;
      
      pts_x2_clone = pts_x2.clone();
      return output, pts_x2_clone;
      
  @staticmethod
  def backward(ctx,grad_output,grad_pts_x2):
      r""" Consider a field u specified at points xj and the mapping to coefficients c at points xi, (uj,xj) --> (ci,xi).
      Computes the gradient of the mapping for backward propagation.
      """
  
      flag_time_it = False;
      if flag_time_it:
        time_1 = time.time();
              
      input,pts_x1,pts_x2 = ctx.saved_tensors;
      
      porder = ctx.atz_porder;
      weight_func = ctx.atz_weight_func;
      weight_func_params = ctx.atz_weight_func_params;      
      cached_data = ctx.atz_cached_data;

      #grad_input = grad_weight_func = grad_weight_func_params = None;
      grad_uj = None;

      # we only compute the gradient in x_i, if it is requested (for efficiency)
      if ctx.needs_input_grad[0]: # derivative in uj
        map_data = cached_data['map_data']; # use cached data
        
        vec_rij = map_data['vec_rij'];
        M_inv   = map_data['M_inv'];

        # c_i = M_{i}^{-1} r_i^T u
        # dF/du = dF/dc*dc/du,
        #
        # We can express this using dF/uj = sum_i dF/dci*dci/duj
        #
        # grad_output = dF/dc,   grad_input = dF/du
        #
        # [grad_input]_j = sum_i dF/ci*dci/duj.
        #
        # In practice, we have both batch and channel indices so
        # grad_output.shape = [batchI,channelI,i,compK]
        # grad_output[batchI,channelI,i,compK] = F(batchI,channelI) with respect to ci[compK](batchI,channelI).
        #
        # grad_input[batchI,channelI,j] =
        #
        # We use matrix broadcasting to get this outcome in practice.
        #

        # @optimize can optimize, since uj only contributes non-zero to a few ci's... and could try to use sparse matrix multiplications.
        A1 = torch.bmm(M_inv,vec_rij); # dci/du, grad = grad[i,compK,j]
        A2 = A1.unsqueeze(0).unsqueeze(0); # match grad_output tensor rank, for grad[batchI,channelI,i,compK,j]
        A3 = grad_output.unsqueeze(4); # grad_output[batchI,channelI,i,compK,j]
        A4 = A3*A2; # elementwise multiplication
        A5 = torch.sum(A4,3); # contract on index compK
        A6 = torch.sum(A5,2); # contract on index i

        grad_uj = A6;
                
      else:
        msg_str = "Requested a currently un-implemented gradient for this map: \n";
        msg_str += "ctx.needs_input_grad = \n" + str(ctx.needs_input_grad); 
        raise Exception(msg_str);

      if flag_time_it:
        msg = 'MapToPoly_Function->backward():';
        msg += 'elapsed_time = %.4e'%(time.time() - time_1);
        print(msg);
    
      return grad_uj,None,None,None,None,None,None,None,None,None,None; # since no trainable parts for these components of map


class MaxPoolOverPoints_Function(torch.autograd.Function):
  r"""Applies a max-pooling operation to obtain values :math:`v_i = \max_{j \in \mathcal{N}_i(\epsilon)} \{u_j\}.` """  
  # @optimize: Should cache the points and neighbor lists.
  #            Should avoid numpy calculations, maybe cache numpy copy of data if needed to avoid .cpu() transfer calls.
  #            Use batching over points to do solves, then GPU parallizable and faster.
  @staticmethod
  def forward(ctx,input,pts_x1,epsilon=None,pts_x2=None,
              indices_xj_i_cache=None,tree_points=None,
              flag_verbose=0):
      r"""Compute max pool operation from values at points (uj,xj) to obtain (vi,xi). 

          Args:
            input (Tensor): The uj values at the location of points xj.
            pts_x1 (Tensor): The collection of domain points :math:`x_j`.
            epsilon (float): The :math:`\epsilon`-neighborhood size to use to sort points (should be compatible with choice of weight_func_params).
            pts_x2 (Tensor): The collection of target points :math:`x_i`.
            tree_points (dict): Stored data to help speed up repeated calculations.
            flag_verbose (int): Level of reporting on progress during the calculations.
            
          Returns:
            tuple: The collection ui at target points (same size as uj in the non-j indices).  The collection xi of target points. Tuple of form (ui,xi).
            
          Note:
            We assume that all inputs are pytorch tensors with pts_x1.shape = [num_pts,num_dim] and similarly for pts_x2.

      """

      ctx.atz_name = 'MaxPoolOverPoints_Function';
      
      ctx.save_for_backward(input,pts_x1,pts_x2);
              
      u = input.clone(); # map input values u(xj) at xj to max value in epsilon neighborhood to u(xi) at xi points.

    	# Assumes that input is of size [k1,k2,...,kn,j], where k1,...,kn are any indices.
    	# We perform maxing over batch over all non-indices in j.
      # We reshape tensor to the form [*,j] where one index in *=index(k1,...,kn).     
      u_num_dim = u.dim();            
      size_nm1  = 1;
      for d in range(u_num_dim - 1): 
        size_nm1 *= u.shape[d];

      uj = u.contiguous().view((size_nm1,u.shape[-1]));  # reshape so indices --> [I,j], I = index(k1,...,kn).

      # reshaped        
      if pts_x2 is None:
        pts_x2 = pts_x1;

      pts_x1_numpy = pts_x1.cpu().numpy(); pts_x2_numpy = pts_x2.cpu().numpy(); # move to cpu to get numpy data
      pts_x1 = pts_x1.to(input.device); pts_x2 = pts_x2.to(input.device); # push back to GPU [@optimize later]        
      num_pts1 = pts_x1.size()[0]; num_pts2 = pts_x2.size()[0];
      if flag_verbose > 0: 
        print("num_pts1 = " + str(num_pts1) + ", num_pts2 = " + str(num_pts2));

      if epsilon is None:
        raise Exception('The epsilon ball size to use around xi must be specified.');
      
      ctx.atz_epsilon = epsilon;

      if indices_xj_i_cache is None:
        flag_need_indices_xj_i = True;
      else:
        flag_need_indices_xj_i = False;

      if flag_need_indices_xj_i and tree_points is None: # build kd-tree of points for neighbor listing            
        tree_points = spatial.cKDTree(pts_x1_numpy);
                  
      ctx.atz_tree_points = tree_points;
      ctx.indices_xj_i_cache = indices_xj_i_cache;

      # Maps from u(x_j) on $x_j \in \mathcal{S}^1$ to a u(x_i) giving max values in epsilon neighborhoods.
      # @optimize by caching these data structure for re-use later
      ui = torch.zeros(size_nm1,num_pts2,requires_grad=False,device=input.device);
      ui_argmax_j = torch.zeros(size_nm1,num_pts2,dtype=torch.int64,requires_grad=False,device=input.device);
      # assumes array of form [*,num_pts2], will be reshaped to match uj, [*,num_pts2].
        
      for i in np.arange(0,num_pts2): # loop over the points $x_i$
        if flag_verbose > 1: print("i = " + str(i) + " : num_pts2 = " + str(num_pts2));

        # find all points distance epsilon from xi
        if flag_need_indices_xj_i:
          indices_xj_i = tree_points.query_ball_point(pts_x2_numpy[i,:], epsilon);
          indices_xj_i = torch.Tensor(indices_xj_i).long();
          indices_xj_i.to(uj.device);
        else:
          indices_xj_i = indices_xj_i_cache[i,:];  # @optimize should consider replacing with better data structures

        # take max over neighborhood. Assumes for now that ui is scalar.
        uuj      = uj[:,indices_xj_i]; 
        qq       = torch.max(uuj,dim=-1,keepdim=True);
        ui[:,i]  = qq[0].squeeze(-1); # store max value
        jj       = qq[1].squeeze(-1); # store index of max value
        ui_argmax_j[:,i] = indices_xj_i[jj]; # store global index of the max value
                  
      # reshape the tensor from ui[I,i] to the form uui[k1,k2,...kn,i]
      uui = ui.view(*u.shape[0:u_num_dim-1],num_pts2);
      uui_argmax_j = ui_argmax_j.view(*u.shape[0:u_num_dim-1],num_pts2);
      ctx.atz_uui_argmax_j = uui_argmax_j; # save for gradient calculation

      output = uui; # for now, we assume for now that ui is scalar array of size [num_pts2]
      output = output.to(input.device);

      return output, pts_x2.clone();

  @staticmethod
  def backward(ctx,grad_output,grad_pts_x2):
      r"""Compute gradients of the max pool operations from values at points (uj,xj) --> (max_ui,xi). """

      flag_time_it = False;
      if flag_time_it:
        time_11 = time.time();
      
      # Compute df/dx from df/dy using the Chain Rule df/dx = df/dx*dy/dx.
      # Compute the gradient with respect to inputs, dz/dx.
      #
      # Consider z = f(g(x)), where we refer to x as the inputs and y = g(x) as outputs.
      # If we know dz/dy, we would like to compute dz/dx.  This will follow from the chain-rule
      # as dz/dx = (dz/dy)*(dy/dx).  We call dz/dy the gradient with respect to output and we call
      # dy/dx the gradient with respect to input.
      #
      # Note: the grad_output can be larger than the size of the input vector if we include in our 
      # definition of gradient_input the derivatives with respect to weights.  Should think of everything
      # input as tilde_x = [x,weights,bias,etc...], then grad_output = dz/dtilde_x.
              
      input,pts_x1,pts_x2 = ctx.saved_tensors;
      
      uui_argmax_j = ctx.atz_uui_argmax_j;

      #grad_input = grad_weight_func = grad_weight_func_params = None;
      grad_input = None;

      # We only compute the gradient in xi, if it is requested (for efficiency)
      # stubs for later possible use, but not needed for now
      if ctx.needs_input_grad[1] or ctx.needs_input_grad[2]:
        msg_str = "Currently requested a non-trainable gradient for this map: \n";
        msg_str += "ctx.needs_input_grad = \n" + str(ctx.needs_input_grad); 
        raise Exception(msg_str);

      if ctx.needs_input_grad[0]:            
        # Compute dL/duj = (dL/dvi)*(dvi/duj), here vi = uui.
        # For the max-pool case, notice that dvi/duj is non-zero only when the index uj 
        # was the maximum value in the neighborhood of vi.  Notice subtle issue with
        # right and left derivatives being different, so max is not differentiable for ties.
        # We use the right derivative lim_h (q(x + h) - q(x))/h, here.

        # We assume that uj.size = [k1,k2,...,kn,j], ui.size = [k1,k2,...,kn,i].
        # These are reshaped so that uuj.size = [I,j] and uui.size = [I,i].
        input_dim    = input.dim();
        size_uj      = input.size();
        size_uj_nm1  = np.prod(size_uj[0:input_dim-1]); # exclude last index size
        
        #ss_grad_input  = input.new_zeros(size_uj_nm1,size_uj[-1]); # to store dL/duj, [I,j] indexing.
        ss_grad_output = grad_output.contiguous().view((size_uj_nm1,grad_output.shape[-1])); # reshape so index [I,i].
        ss_uui_argmax_j = uui_argmax_j.contiguous().view((size_uj_nm1,grad_output.shape[-1])); # reshape so index [I,i].

        # assign the entries k_i = argmax_{j in Omega_i} uj, reshaped so [*,j] = val[*,j].
        flag_method = 'method1';
        if flag_method == 'method1':

          flag_time_it = False;
          if flag_time_it:
            time_0 = time.time();

          I = torch.arange(0,size_uj_nm1,dtype=torch.int64,device=input.device);
          vec_ones = torch.ones(grad_output.shape[-1],dtype=torch.int64,device=input.device);
          II = torch.ger(I.float(),vec_ones.float()); # careful int --> float conv
          II = II.flatten();
          JJ = ss_uui_argmax_j.flatten();
          IJ_indices1 = torch.stack((II,JJ.float())).long();
                      
          i_index  = torch.arange(0,grad_output.shape[-1],dtype=torch.int64,device=input.device);
          vec_ones = torch.ones(size_uj_nm1,dtype=torch.int64,device=input.device);
          KK       = torch.ger(vec_ones.float(),i_index.float()); # careful int --> float conv
          KK       = KK.flatten();
          IJ_indices2 = torch.stack((II,KK)).long();

          # We aim to compute dL/duj = dL/d\bar{u}_i*d\bar{u}_i/duj.
          #
          # This is done efficiently by constructing a sparse matrix using how \bar{u}_i
          # depends on the uj.  Sometimes the same uj contributes multiple times to
          # a given \bar{u}_i entry, so we add together those contributions, as would
          # occur in an explicit multiplication of the terms above for dL/duj.
          # This is acheived efficiently using the .add() for sparse tensors in PyTorch.
          
          # We construct entries of the sparse matrix and coelesce them (add repeats).
          vals = ss_grad_output[IJ_indices2[0,:],IJ_indices2[1,:]]; # @optimize, maybe just flatten
          N1 = size_uj_nm1; N2 = size_uj[-1]; sz = torch.Size([N1,N2]);
          ss_grad_input = torch.sparse.FloatTensor(IJ_indices1,vals,sz).coalesce().to_dense();

          if flag_time_it:
            time_1 = time.time();
            
            print("time: backward(): compute ss_grad_input = %.4e sec"%(time_1 - time_0));            

        elif flag_method == 'method2':
          II      = torch.arange(0,size_uj_nm1,dtype=torch.int64);
          i_index = torch.arange(0,grad_output.shape[-1],dtype=torch.int64);
        
          # @optimize by vectorizing this calculation
          for I in II:
            for j in range(0,i_index.shape[0]):
              ss_grad_input[I,ss_uui_argmax_j[I,j]] += ss_grad_output[I,i_index[j]];
            
        else:
          raise Exception("flag_method type not recognized.\n flag_method = %s"%flag_method);    
          
        # reshape
        grad_input = ss_grad_input.view(*size_uj[0:input_dim - 1],size_uj[-1]);

      if flag_time_it:
        msg = 'atzGMLS_MaxPool2D_Function->backward(): ';
        msg += 'elapsed_time = %.4e'%(time.time() - time_11);
        print(msg);  
      
      return grad_input,None,None,None,None,None,None; # since no trainable parts for components of this map

class ExtractFromTuple_Function(torch.autograd.Function):
  r"""Extracts from a tuple of outputs one of the components."""
    
  @staticmethod
  def forward(ctx,input,index):
    r"""Extracts tuple entry with the specified index."""
    ctx.atz_name = 'ExtractFromTuple_Function';
    
    extracted = input[index];
    output = extracted.clone(); # clone added for safety
    
    return output;
    
  @staticmethod
  def backward(ctx,grad_output): # number grad's needs to match outputs of forward
    r"""Computes gradient of the extraction."""

    raise Exception('This backward is not implemented, since PyTorch automatically handled this in the past.');
    return None,None;
        
# ====================================
# Custom Modules
# ====================================
class PdbSetTraceLayer(nn.Module):
  r"""Allows for placing break-points within the call sequence of layers using pdb.set_trace().  Helpful for debugging networks."""

  def __init__(self):
    r"""Initialization (currently nothing to do, but call super-class)."""
    super(PdbSetTraceLayer, self).__init__()
            
  def forward(self, input):            
    r"""Executes a PDB breakpoint inside of a running network to help with debugging."""
    out = input.clone();  # added clone to avoid .grad_fn overwrite
    pdb.set_trace();
    return out;

class ExtractFromTuple(nn.Module): 
  r"""Extracts from a tuple of outputs one of the components."""
     
  def __init__(self,index=0):
    r"""Initializes the index to extract."""
    super(ExtractFromTuple, self).__init__()
    self.index     = index;  

  def forward(self, input):
    r"""Extracts the tuple entry with the specified index."""
    extracted = input[self.index];
    extracted_clone = extracted.clone(); # cloned to avoid overwrite of .grad_fn
    return extracted_clone;

class ReshapeLayer(nn.Module):
  r"""Performs reshaping of a tensor output within a network."""
 
  def __init__(self,reshape,permute=None):
    r"""Initializes the reshaping form to use followed by the indexing permulation to apply."""
    super(ReshapeLayer, self).__init__()
    self.reshape = reshape; 
    self.permute = permute;
    
  def forward(self, input):        
    r"""Reshapes the tensor followed by applying a permutation to the indexing."""
    reshape = self.reshape;
    permute = self.permute;  
    A = input.contiguous();
    out = A.view(*reshape); 
    if permute is not None:
      out = out.permute(*permute);
    return out;

class PermuteLayer(nn.Module):
  r"""Performs permutation of indices of a tensor output within a network."""
   
  def __init__(self,permute=None):
    r"""Initializes the indexing permuation to apply to tensors."""
    super(PermuteLayer, self).__init__()
    self.permute = permute;
    
  def forward(self, input):            
    r"""Applies and indexing permuation to the input tensor."""
    permute = self.permute;
    input_clone = input.clone(); # adding clone to avoid .grad_fn overwrites
    out = input_clone.permute(*permute);
    return out;

class MLP_Pointwise(nn.Module):
  r"""Creates a collection of multilayer perceptrons (MLPs) for each output channel.
  The MLPs are then applied at each target point xi.
  """

  def create_mlp_unit(self,layer_sizes,unit_name='',flag_bias=True):
    r"""Creates an instance of an MLP with specified layer sizes. """
    layer_dict = OrderedDict();
    NN = len(layer_sizes);
    for i in range(NN - 1):
      key_str = unit_name + ':hidden_layer_%.4d'%(i + 1);
      layer_dict[key_str] = nn.Linear(layer_sizes[i], layer_sizes[i+1],bias=flag_bias);
      if i < NN - 2: # last layer should be linear
        key_str = unit_name + ':relu_%.4d'%(i + 1);
        layer_dict[key_str] = nn.ReLU();
        
    mlp_unit = nn.Sequential(layer_dict); # uses ordered dictionary to create network

    return mlp_unit;
  
  def __init__(self,layer_sizes,channels_in=1,channels_out=1,flag_bias=True,flag_verbose=0):
    r"""Initializes the structure of the pointwise MLP module with layer sizes, number input channels, number of output channels.

        Args:
          layer_sizes (list): The number of hidden units in each layer.
          channels_in (int): The number of input channels.
          channels_out (int): The number of output channels.
          flag_bias (bool): If the MLP should include the additive bias b added into layers.
          flag_verbose (int): The level of messages generated on progress of the calculation.

    """
    super(MLP_Pointwise, self).__init__();

    self.layer_sizes   = layer_sizes; 
    self.flag_bias     = flag_bias;
    self.depth         = len(layer_sizes); 

    self.channels_in   = channels_in;
    self.channels_out  = channels_out;
        
    # create intermediate layers
    mlp_list = nn.ModuleList();
    layer_sizes_unit = layer_sizes.copy(); # we use inputs k*c to cross channels in practice in our unit MLPs
    layer_sizes_unit[0] = layer_sizes_unit[0]*channels_in; # modify the input to have proper size combined k*c
    for ell in range(channels_out):
      mlp_unit = self.create_mlp_unit(layer_sizes_unit,'unit_ell_%.4d'%ell,flag_bias=flag_bias);
      mlp_list.append(mlp_unit);

    self.mlp_list = mlp_list;  
      
  def forward(self, input, params = None): 
    r"""Applies the specified MLP pointwise to the collection of input data to produce pointwise entries of the output channels."""
    #
    # Assumes the tensor has the form [i1,i2,...in,k,c], the last two indices are the
    # channel index k, and the coefficient index c, combine for ease of use, but can reshape.
    # We collapse input tensor with indexing [i1,i2,...in,k,c] to a [I,k*c] tensor, where
    # I is general index, k is channel, and c are coefficient index.
    #
    s = input.shape;
    num_dim = input.dim();

    if (s[-2] != self.channels_in) or (s[-1] != self.layer_sizes[0]): # check correct sized inputs
      print("input.shape = " + str(input.shape));
      raise Exception("MLP assumes an input tensor of size [*,%d,%d]"%(self.channels_in,self.layer_sizes[0]));
    
    calc_size1 = 1.0;
    for d in range(num_dim-2):
      calc_size1 *= s[d];
    calc_size1 = int(calc_size1);
    
    x = input.contiguous().view(calc_size1,s[num_dim-2]*s[num_dim-1]); # shape input to have indexing [I,k*NN + c]

    if params is None:
      output = torch.zeros((self.channels_out,x.shape[0]),device=input.device); # shape [ell,*]
      for ell in range(self.channels_out):
        mlp_q = self.mlp_list[ell];  
        output[ell,:] = mlp_q.forward(x).squeeze(-1); # reduce from [N,1] to [N]

      s = input.shape;
      output = output.view(self.channels_out,*s[0:num_dim-2]); # shape to have index [ell,i1,i2,...,in]
      nn = output.dim();       
      p_ind = np.arange(nn) + 1;
      p_ind[nn-1] = 0;
      p_ind = tuple(p_ind);
      output = output.permute(p_ind); # [*,ell] indexing of final shape
    else:
      raise Exception("Not yet implemented for setting parameters.");

    return output; # [*,ell] indexing of final shape

  def to(self, device):
    r"""Moves data to GPU or other specified device."""
    super(MLP_Pointwise, self).to(device);
    for ell in range(self.channels_out):
      mlp_q = self.mlp_list[ell];
      mlp_q.to(device);      
    return self;


class MLP1(nn.Module):
  r"""Creates a multilayer perceptron (MLP). """

  def __init__(self, layer_sizes, flag_bias = True, flag_verbose=0):
    r"""Initializes MLP and specified layer sizes."""
    super(MLP1, self).__init__();

    self.layer_sizes   = layer_sizes; 
    self.flag_bias     = flag_bias;
    self.depth         = len(layer_sizes); 
    
    # create intermediate layers
    layer_dict = OrderedDict();
    NN = len(layer_sizes);
    for i in range(NN - 1):
      key_str = 'hidden_layer_%.4d'%(i + 1);
      layer_dict[key_str] = nn.Linear(layer_sizes[i], layer_sizes[i+1],bias=flag_bias);
      if i < NN - 2: # last layer should be linear
        key_str = 'relu_%.4d'%(i + 1);
        layer_dict[key_str] = nn.ReLU();
        
    self.layers = nn.Sequential(layer_dict); # uses ordered dictionary to create network
          
  def forward(self, input, params = None): 
    r"""Applies the MLP to the input data.
     
        Args:
          input (Tensor): The coefficient channel data organized as one stacked 
          vector of size Nc*M, where Nc is number of channels and M is number of 
          coefficients per channel.

        Returns:
          Tensor: The evaluation of the network.  Returns tensor of size [batch,1].

    """
    # evaluate network with specified layers
    if params is None:
      eval = self.layers.forward(input);
    else:
      raise Exception("Not yet implemented for setting parameters.");
      
    return eval;

  def to(self, device):
    r"""Moves data to GPU or other specified device."""
    super(MLP1, self).to(device);
    self.layers = self.layers.to(device);
    return self;

class MapToPoly(nn.Module):
  r"""
  This layer processes a collection of scattered data points consisting of a collection 
  of values :math:`u_j` at points :math:`x_j`.  For a collection of target points 
  :math:`x_i`, local least-squares problems are solved for obtaining a local representation 
  of the data over a polynomial space.  The layer outputs a collection of polynomial 
  coefficients :math:`c(x_i)` at each point and the collection of target points :math:`x_i`.
  """
  def __init__(self, porder, weight_func, weight_func_params, pts_x1, 
               epsilon = None,pts_x2 = None,tree_points = None,
               device = None,flag_verbose = 0,**extra_params):
      r"""Initializes the layer for mapping between field data uj at points xj to the 
          local polynomial reconstruction represented by coefficients ci at target points xi.

          Args:
            porder (int): Order of the basis to use.  For polynomial basis is the degree.
            weight_func (func): Weight function to use.
            weight_func_params (dict): Weight function parameters.
            pts_x1 (Tensor): The collection of domain points :math:`x_j`.
            epsilon (float): The :math:`\epsilon`-neighborhood size to use to sort points (should be compatible with choice of weight_func_params).
            pts_x2 (Tensor): The collection of target points :math:`x_i`.
            tree_points (dict): Stored data to help speed up repeated calculations.
            device: Device on which to perform calculations (GPU or other, default is CPU).
            flag_verbose (int): Level of reporting on progress during the calculations.
            **extra_params: Extra parameters allowing for specifying layer name and caching mode.            

      """
      super(MapToPoly, self).__init__();

      self.flag_verbose = flag_verbose;

      if device is None:
        device = torch.device('cpu');

      self.device = device;
      
      if 'name' in extra_params:
        self.name = extra_params['name'];
      else:
        self.name = "default_name";
  
      if 'flag_cache_mode' in extra_params:
        flag_cache_mode = extra_params['flag_cache_mode'];
      else:
        flag_cache_mode = 'generate1';
        
      if flag_cache_mode == 'generate1':  # setup from scratch
        self.porder = porder;
        self.weight_func = weight_func;
        self.weight_func_params = weight_func_params;

        self.pts_x1 = pts_x1;
        self.pts_x2 = pts_x2;

        self.pts_x1_numpy = None;
        self.pts_x2_numpy = None;        
      
        if self.pts_x2 is None:
          self.pts_x2 = pts_x1;

        self.epsilon = epsilon;
      
        if tree_points is None: # build kd-tree of points for neighbor listing
          if self.pts_x1_numpy is None: self.pts_x1_numpy = pts_x1.cpu().numpy();
          self.tree_points = spatial.cKDTree(self.pts_x1_numpy);

        if device is None:
          device = torch.device('cpu');

        self.device = device;
                  
        self.cached_data = {}; # create empty cache for storing data          
        generate_mapping = MapToPoly_Function.generate_mapping;
        self.cached_data['map_data'] = generate_mapping(self.weight_func,self.weight_func_params,
                                                        self.porder,self.epsilon, 
                                                        self.pts_x1,self.pts_x2,
                                                        self.tree_points,self.device,
                                                        self.flag_verbose);
        
      elif flag_cache_mode == 'load_from_file': # setup by loading data from cache file
          
        if 'cache_filename' in extra_params:
          cache_filename = extra_params['cache_filename'];
        else:
          raise Exception('No cache_filename specified.');

        self.load_cache_data(cache_filename); # load data from file
          
      else:
        print("flag_cache_mode = " + str(flag_cache_mode));
        raise Exception('flag_cache_mode is invalid.');

  def save_cache_data(self,cache_filename):
    r"""Save needed matrices and related data to .pickle for later cached use. (Warning: prototype codes here currently and not tested)."""
    # collect the data to save  
    d = {};
    d['porder']  = self.porder;      
    d['epsilon'] = self.epsilon;
    if self.pts_x1_numpy is None: self.pts_x1_numpy = pts_x1.cpu().numpy();
    d['pts_x1'] = self.pts_x1_numpy;
    if self.pts_x2_numpy is None: self.pts_x2_numpy = pts_x2.cpu().numpy();            
    d['pts_x2'] = self.pts_x2_numpy;
    d['weight_func_str'] = str(self.weight_func);
    d['weight_func_params'] = self.weight_func_params;
    d['version'] = __version__;  # Module version      
    d['cached_data'] = self.cached_data;

    # write the data to disk
    f = open(cache_filename,'wb');
    p.dump(d,f); # load the data from file
    f.close();

  def load_cache_data(self,cache_filename):
      r"""Load the needed matrices and related data from .pickle. (Warning: prototype codes here currently and not tested)."""
      f = open(cache_filename,'rb');
      d = p.load(f); # load the data from file
      f.close();

      print(d.keys())
      self.porder = d['porder'];
      self.epsilon = d['epsilon'];

      self.weight_func = d['weight_func_str'];
      self.weight_func_params = d['weight_func_params'];

      self.pts_x1 = torch.from_numpy(d['pts_x1']).to(device);
      self.pts_x2 = torch.from_numpy(d['pts_x2']).to(device);

      self.pts_x1_numpy = d['pts_x1'];
      self.pts_x2_numpy = d['pts_x2'];

      if self.pts_x2 is None:
        self.pts_x2 = pts_x1;

      # build kd-tree of points for neighbor listing
      if self.pts_x1_numpy is None: self.pts_x1_numpy = pts_x1.cpu().numpy();
      self.tree_points = spatial.cKDTree(self.pts_x1_numpy);

      self.cached_data = d['cached_data'];
    
  def eval_poly(self,pts_x,pts_x2_i0,c_star_i0,porder=None,flag_verbose=None):
    r"""Evaluates the polynomial reconstruction around a given target point pts_x2_i0."""
    if porder is None:
      porder = self.porder;
    if flag_verbose is None:
      flag_verbose = self.flag_verbose;
    MapToPoly_Function.eval_poly(pts_x,pts_x2_i0,c_star_i0,porder,flag_verbose);
      
  def forward(self, input): # define the action of this layer
      r"""For a field u specified at points xj, performs the mapping to coefficients c at points xi, (uj,xj) :math:`\rightarrow` (ci,xi)."""
      flag_time_it = False;
      if flag_time_it:
        time_1 = time.time();
        
      # We evaluate the action of the function, backward will be called automatically when computing gradients.
      uj = input;
      output = MapToPoly_Function.apply(uj,self.porder,
                                            self.weight_func,self.weight_func_params,
                                            self.pts_x1,self.epsilon,self.pts_x2,
                                            self.cached_data,self.tree_points,self.device,
                                            self.flag_verbose);
      if flag_time_it:
        msg = 'MapToPoly->forward(): ';
        msg += 'elapsed_time = %.4e'%(time.time() - time_1);
        print(msg);
        
      return output;
  
  def extra_repr(self):
      r"""Displays information associated with this module."""
      # Display some extra information about this layer.
      return 'porder={}, weight_func={}, weight_func_params={}, pts_x1={}, pts_x2={}'.format(
          self.porder, self.weight_func, self.weight_func_params, self.pts_x1.shape, self.pts_x2.shape
      );

  def to(self, device):
    r"""Moves data to GPU or other specified device."""
    super(MapToPoly,self).to(device);
    self.pts_x1 = self.pts_x1.to(device);
    self.pts_x2 = self.pts_x2.to(device);
    return self;  

class MaxPoolOverPoints(nn.Module):
  r"""Applies a max-pooling operation to obtain values :math:`v_i = \max_{j \in \mathcal{N}_i(\epsilon)} \{u_j\}.` """  
  def __init__(self,pts_x1,epsilon=None,pts_x2=None,
               indices_xj_i_cache=None,tree_points=None,
               device=None,flag_verbose=0,**extra_params):
      r"""Setup of max-pooling operation.

          Args:
            pts_x1 (Tensor): The collection of domain points :math:`x_j`.  We assume size [num_pts,num_dim].
            epsilon (float): The :math:`\epsilon`-neighborhood size to use to sort points (should be compatible with choice of weight_func_params).
            pts_x2 (Tensor): The collection of target points :math:`x_i`.
            indices_xj_i_cache (dict): Stored data to help speed up repeated calculations.
            tree_points (dict): Stored data to help speed up repeated calculations.
            device: Device on which to perform calculations (GPU or other, default is CPU).
            flag_verbose (int): Level of reporting on progress during the calculations.
            **extra_params (dict): Extra parameters allowing for specifying layer name and caching mode. 
       
      """      
      super(MaxPoolOverPoints,self).__init__();

      self.flag_verbose = flag_verbose;

      if device is None:
        device = torch.device('cpu');

      self.device = device;
      
      if 'name' in extra_params:
        self.name = extra_params['name'];
      else:
        self.name = "default_name";
  
      if 'flag_cache_mode' in extra_params:
        flag_cache_mode = extra_params['flag_cache_mode'];
      else:
        flag_cache_mode = 'generate1';

      if flag_cache_mode == 'generate1':  # setup from scratch
        self.pts_x1 = pts_x1;
        self.pts_x2 = pts_x2;

        self.pts_x1_numpy = None;
        self.pts_x2_numpy = None;        
      
        if self.pts_x2 is None:
          self.pts_x2 = pts_x1;

        self.epsilon = epsilon;
      
        if tree_points is None: # build kd-tree of points for neighbor listing
          if self.pts_x1_numpy is None: self.pts_x1_numpy = pts_x1.cpu().numpy();
          self.tree_points = spatial.cKDTree(self.pts_x1_numpy);

        if indices_xj_i_cache is None:
          self.indices_xj_i_cache = None; # cache the neighbor lists around each xi
        else:
          self.indices_xj_i_cache = indices_xj_i_cache;

        if device is None:
          device = torch.device('cpu');

        self.device = device;
                  
        self.cached_data = {}; # create empty cache for storing data
        
      elif flag_cache_mode == 'load_from_file': # setup by loading data from cache file
          
        if 'cache_filename' in extra_params:
          cache_filename = extra_params['cache_filename'];
        else:
          raise Exception('No cache_filename specified.');

        self.load_cache_data(cache_filename); # load data from file
          
      else:
        print("flag_cache_mode = " + str(flag_cache_mode));
        raise Exception('flag_cache_mode is invalid.');

  def save_cache_data(self,cache_filename):
    r"""Save data to .pickle file for caching. (Warning: Prototype placeholder code.)"""      
    # collect the data to save  
    d = {};
    d['epsilon'] = self.epsilon;
    if self.pts_x1_numpy is None: self.pts_x1_numpy = pts_x1.cpu().numpy();
    d['pts_x1'] = self.pts_x1_numpy;
    if self.pts_x2_numpy is None: self.pts_x2_numpy = pts_x2.cpu().numpy();            
    d['pts_x2'] = self.pts_x2_numpy;
    d['version'] = __version__;  # Module version      
    d['cached_data'] = self.cached_data;

    # write the data to disk
    f = open(cache_filename,'wb');
    p.dump(d,f); # load the data from file
    f.close();

  def load_cache_data(self,cache_filename):
      r"""Load data to .pickle file for caching. (Warning: Prototype placeholder code.)"""      
      f = open(cache_filename,'rb');
      d = p.load(f); # load the data from file
      f.close();

      print(d.keys())
      self.epsilon = d['epsilon'];

      self.pts_x1 = torch.from_numpy(d['pts_x1']).to(device);
      self.pts_x2 = torch.from_numpy(d['pts_x2']).to(device);

      self.pts_x1_numpy = d['pts_x1'];
      self.pts_x2_numpy = d['pts_x2'];

      if self.pts_x2 is None:
        self.pts_x2 = pts_x1;

      # build kd-tree of points for neighbor listing
      if self.pts_x1_numpy is None: self.pts_x1_numpy = pts_x1.cpu().numpy();
      self.tree_points = spatial.cKDTree(self.pts_x1_numpy);

      self.cached_data = d['cached_data'];
            
  def forward(self, input): # define the action of this layer
      r"""Applies a max-pooling operation to obtain values :math:`v_i = \max_{j \in \mathcal{N}_i(\epsilon)} \{u_j\}.` 

          Args:
            input (Tensor): The collection uj of field values at the points xj.

          Returns:
            Tensor: The collection of field values vi at the target points xi.

      """      
      flag_time_it = False;
      if flag_time_it:
        time_1 = time.time();
      
      uj = input;
      output = MaxPoolOverPoints_Function.apply(uj,self.pts_x1,self.epsilon,self.pts_x2,
                                                self.indices_xj_i_cache,self.tree_points,
                                                self.flag_verbose);
      
      if flag_time_it:
        msg = 'MaxPoolOverPoints->forward(): ';
        msg += 'elapsed_time = %.4e'%(time.time() - time_1);
        print(msg);

      return output;

  def extra_repr(self):
      r"""Displays information associated with this module."""
      return 'pts_x1={}, pts_x2={}'.format(self.pts_x1.shape, self.pts_x2.shape);

  def to(self, device):
    r"""Moves data to GPU or other specified device."""
    super(MaxPoolOverPoints, self).to(device);
    self.pts_x1 = self.pts_x1.to(device);
    self.pts_x2 = self.pts_x2.to(device);
    return self;

class GMLS_Layer(nn.Module):
  r"""The GMLS-Layer processes scattered data by using Generalized Moving Least 
  Squares (GMLS) to construct a local reconstruction of the data (here polynomials).  
  This is represented by coefficients that are mapped to approximate the action of 
  linear or non-linear operators on the input field.

  As depicted above, the architecture processes a collection of input channels 
  into intermediate coefficient channels.  The coefficient channels are 
  then collectively mapped to output channels.  The mappings can be any unit 
  for which back-propagation can be performed.  This includes linear 
  layers or non-linear maps based on multilayer perceptrons (MLPs).

  Examples:

    Here is a typical way to construct a GMLS-Layer.  This is done in 
    the following stages.

    ``(i)`` Construct the scattered data locations xj, xi at which processing will occur.  Here, we create points in 2D.

    >>> xj = torch.randn((100,2),device=device); xi = torch.randn((100,2),device=device);
 
    ``(ii)`` Construct the mapping unit that will be applied pointwise.  Here we create an MLP 
    with Nc input coefficient channels and channels_out output channels.

    >>> layer_sizes = [];
    >>> num_input   = Nc*num_polys; # number of channels (NC) X number polynomials (num_polys) (cross-channel coupling allowed)
    >>> num_depth   = 4; num_hidden  = 100; channels_out = 16; # depth, width, number of output filters 
    >>> layer_sizes.append(num_polys);
    >>> for k in range(num_depth):
    >>>   layer_sizes.append(num_hidden);
    >>> layer_sizes.append(1); # a single unit always gives scalar output, we then use channels_out units.    
    >>> mlp_q_map1 = gmlsnets_pytorch.nn.MLP_Pointwise(layer_sizes,channels_out=channels_out);

    ``(iii)`` Create the GMLS-Layer using these components.

    >>> weight_func1 = gmlsnets_pytorch.nn.MapToPoly_Function.weight_one_minus_r; 
    >>> weight_func_params = {'epsilon':1e-3,'p'=4};
    >>> gmls_layer_params = {
              'flag_case':'standard','porder':4,'Nc':3,
              'mlp_q1':mlp_q_map1,
              'pts_x1':xj,'pts_x2':xi,'epsilon':1e-3,
              'weight_func1':weight_func1,'weight_func1_params':weight_func1_params,
              'device':device,'flag_verbose':0
              };
    >>> gmls_layer=gmlsnets_pytorch.nn.GMLS_Layer(**gmls_layer_params);

    Here is an example of how a GMLS-Layer and other modules in this package
    can be used to process scattered data.  This could be part of a larger 
    neural network in practice (see example codes for more information).  For instance,
        
    >>> layer1 = nn.Sequential(gmls_layer, # produces output tuple of tensors (ci,xi) with shapes ([batch,ci,xi],[xi]).
                               #PdbSetTraceLayer(),
                               ExtractFromTuple(index=0), # from output keep only the ui part and discard the xi part.
                               #PdbSetTraceLayer(),
                               PermuteLayer((0,2,1)) # organize indexing to be [batch,xi,ci], for further processing.
                               ).to(device);

    You can uncomment the PdbSetTraceLayer() to get breakpoints for state information and tensor shapes during processing.
    The PermuteLayer() changes the order of the indexing.  Also can use ReshapeLayer() to reshape the tensors, which is 
    especially useful for processing data related to CNNs.

    Much of the construction can be further simplified by writing a few wrapper classes for your most common use cases.

  More information also can be found in the example codes directory.
  """  
  def __init__(self, flag_case, porder, pts_x1, epsilon, weight_func, weight_func_params,
               mlp_q = None,pts_x2 = None, device = None, flag_verbose = 0):
    r"""
        Initializes the GMLS layer.  

        Args:
          flag_case (str): Flag for the type of architecture to use (default is 'standard').
          porder (int): Order of the basis to use (polynomial degree).
          pts_x1 (Tensor): The collection of domain points :math:`x_j`.
          epsilon (float): The :math:`\epsilon`-neighborhood size to use to sort points (should be compatible with choice of weight_func_params).
          weight_func (func): Weight function to use.
          weight_func_params (dict): Weight function parameters.
          mlp_q (module): Mapping q unit for computing :math:`q(c)`, where c are the coefficients.
          pts_x2 (Tensor): The collection of target points :math:`x_i`.
          device: Device on which to perform calculations (GPU or other, default is CPU).
          flag_verbose (int): Level of reporting on progress during the calculations.
    """
    super(GMLS_Layer, self).__init__();

    if flag_case is None:
      self.flag_case = 'standard';
    else:
      self.flag_case = flag_case;
    
    if device is None:
      device = torch.device('cpu');
    
    self.device = device;

    if self.flag_case == 'standard':
      tree_points = None;
      self.MapToPoly_1 = MapToPoly(porder, weight_func, weight_func_params, 
                                                                      pts_x1, epsilon, pts_x2, tree_points,
                                                                      device, flag_verbose);
      
      if mlp_q is None: # if not specified then create some default custom layers
        raise Exception("Need to specify the mlp_q module for mapping coefficients to output.");    
      else: # in this case initialized outside
        self.mlp_q = mlp_q;

    else:
      print("flag_case = " + str(flag_case));        
      print("self.flag_case = " + str(self.flag_case));
      raise Exception('flag_case not valid.');
      
  def forward(self, input):
    r"""Computes GMLS-Layer processing scattered data input field uj to obtain output field vi.

        Args: 
          input (Tensor): Input channels uj organized in the shape [batch,xj,uj].    

        Returns:
          tuple: The output channels and point locations (vi,xi).  The field vi = q(ci).

    """
    if self.flag_case == 'standard':

      map_output     = self.MapToPoly_1.forward(input); 
      c_star_i       = map_output[0];
      pts_x2         = map_output[1];
      
      # MLP should apply across all channels and coefficients (coeff capture spatial, like kernel)
      fc_input = c_star_i.permute((0,2,1,3)); # we organize as [batchI,ptsI,channelsI,coeffI]

      # We assume MLP can process channelI*Nc + coeffI.
      # We assume output of out = fc, has shape [batchI,ptsI,channelsNew]
      # Outside routines can reshape that into an nD array again for structure samples or use over scattered samples.
      q_of_c_star_i  = self.mlp_q.forward(fc_input);
                                                                     
      pts_x2_p = None; # currently returns None to simplify back-prop and debugging, but could just return the pts_x2.
      
      return_vals = q_of_c_star_i, pts_x2_p;

    return return_vals;

  def to(self, device):
    r"""Moves data to GPU or other specified device."""
    super(GMLS_Layer, self).to(device);
    self.MapToPoly_1 = self.MapToPoly_1.to(device);
    self.mlp_q = self.mlp_q.to(device);
    return self;




"""
  Collection of utility routines.
"""

# Authors: B.J. Gross and P.J. Atzberger
# Website: http://atzberger.org/

import torch;
import torch.nn as nn;
import numpy as np;

import pdb
import time

#******************************************
# Custom Functions
#******************************************
class PeriodicPad2dFunc(torch.autograd.Function):
    """Performs periodic padding/tiling of a 2d lattice."""

    @staticmethod
    def forward(ctx, input, pad=None,flag_coords=False):
        """
        Periodically tiles the input into a 2d array for 3x3 repeat pattern.  
        This allows for a quick way for handling periodic boundary conditions 
        on the units cell.
        
        Args:
          input (Tensor): A tensor of size [nbatch,nchannel,nx,ny] or [nx,ny].
          pad (float): The pad value to use.
          flag_coord (bool): If beginning components are coordinates (x,y,z,...,u).

        We then adjust (x,y,u) --> (x + i, y + j, u) for image (i,j).          
        
        Returns:
          output (Tensor): A tensor of the same size [nbatch,nchannel,3*nx,3*ny] or [3*nx,3*ny].
        
        """
        # We use alternative by concatenating the arrays together to tile.
        # Process Tensor inputs of the shape with [nbatch,nchannel,nx,ny].
        # We also allow the case of [nx,ny] for single input.               
        nd = input.dim();
        if (nd > 4):
          raise Exception('Expects tensor that has number of dimensions <= 4 (dim = {}).'.format(nd));
        
        if (nd < 2):
          raise Exception('Expects tensor that has at least number of dimensions >= 2 (dim = {}).'.format(nd));            
        
        a  = input;
        if (nd == 2):
          # add extra dimensions so we can process all tensors the same way
          a = input.unsqueeze(0).unsqueeze(0); # size --> [1,1,nx,ny]          

        w1 = w2 = a;
        aa = torch.cat([w1,a,w2],dim=2);
        h1 = h2 = aa;
        output = torch.cat([h1,aa,h2],dim=3);

        if flag_coords: # indicates (x,y,u) input, so need
                        # to adjust x + i-1 and y + j-1, i,j = 0,1,2 to extend.
          coordI1 = 0; coordI2 = 1; #x and y components
          N1 = output.shape[2]//3; # block size
          N2 = output.shape[3]//3;
          for j in range(0,3):
            blockI2 = np.arange(N2*j,N2*(j + 1),dtype=int);
            output[:,coordI1,:,blockI2] += (j - 1);

          for i in range(0,3):
            blockI1 = np.arange(N1*i,N1*(i + 1),dtype=int);
            output[:,coordI2,blockI1,:] += (i - 1);              
        
        if (nd == 2): # if only [nx,ny] then squeeze our extra dimensions
          output = output.squeeze(0).squeeze(0);        
        
        ctx.pad = pad;
        ctx.size = input.size();
        ctx.numel = input.numel();
        ctx.num_tiles = 3;
                
        return output;

    @staticmethod
    def backward(ctx, grad_output):
        r"""Compute df/dx from df/dy using the Chain Rule df/dx = df/dx*dy/dx.
            For periodic padding we use df/dx = sum_{y_i ~ x} df/dy_i, where 
            the y_i ~ x are all points y_i that are equivalent to x under the periodic extension.
        """              
        num_tiles = ctx.num_tiles;
        
        nd = input.dim();
        if (nd > 4):
          raise Exception('Expects tensor that has number of dimensions <= 4 (dim = {}).'.format(nd));
        
        if (nd < 2):
          raise Exception('Expects tensor that has at least number of dimensions >= 2 (dim = {}).'.format(nd));                    
        
        b = grad_output; # short-hand for grad_output (size of output)
        if (nd == 2):
          # add extra dimensions so we can process all tensors the same way
          b = b.unsqueeze(0).unsqueeze(0); # size --> [1,1,nx,ny]          
                                       
        # construct indices to contract the periodic images in each dimension
        ind_r = torch.zeros(ctx.size,dtype=torch.int64);
        torch.arange(0, ctx.size[2]*num_tiles, out=ind_r);
        ind_r = ind_r.fmod(ctx.size[2]); # repeat indices number of rows [0,1,..nrow-1,0,...nrow-1].
        ind_r = ind_r.view(-1);

        ind_c = torch.zeros(ctx.size,dtype=torch.int64);
        torch.arange(0, ctx.size[3]*num_tiles, out=ind_c);
        ind_c = ind_c.fmod(ctx.size[3]); # repeat indices number of cols [0,1,..ncol-1,0,...ncol-1].
        ind_c = ind_c.view(-1);

        c = b.new_zeros(ctx.size[0],ctx.size[1],ctx.size[2],ctx.size[3]*num_tiles);

        c = c.index_add(2,ind_r,grad_output); # add the rows together to start contracting periodicity        

        d = b.new_zeros(a.transpose(2,3).size());
        d = d.index_add(2,ind_c,c.t()); # add the cols together to contract periodicity
        grad_input = d.transpose(2,3); # Note, this includes contraction of grad_output already, so is df/dx.
                
        if (nd == 2): # if only [nx,ny] then squeeze out extra dimensions
          output = grad_input.squeeze(0).squeeze(0);

        return grad_input;
            
#------------------
class ExtractUnitCell2dFunc(torch.autograd.Function):
    r"""Extracts the 2d unit cell from periodic tiling."""
    
    @staticmethod
    def forward(ctx, input, pad=None):
        r"""Extracts the 2d unit cell from periodic tiling.
        
        Args:
          input (Tensor): Tensor of size [nbatch,nchannel,nx,ny] or [nx,ny].
          
        Returns:
          Tensor: The Tensor of size [nbatch,nchannel,nx/3,ny/3] or [nx/3,ny/3].
                
        """ 
        num_tiles = 3;
        ctx.num_tiles = num_tiles;
        
        nd = input.dim();
        if (nd > 4):
          raise Exception('Expects tensor that has number of dimensions <= 4 (dim = {}).'.format(nd));
        
        if (nd < 2):
          raise Exception('Expects tensor that has at least number of dimensions >= 2 (dim = {}).'.format(nd));                    
        
        a = input; # short-hand 
        if (nd == 2):
          # add extra dimensions so we can process all tensors the same way
          a = a.unsqueeze(0).unsqueeze(0); # size --> [1,1,nx,ny]
        
        chunks_col = torch.chunk(a,num_tiles,dim=2);
        aa = torch.chunk(chunks_col[1],num_tiles,dim=3);  # extract middle of middle
        
        output = aa[1]; # choose middle one (assumes num_tiles==3)
                        
        if (nd == 2): # if only [nx,ny] then squeeze out extra dimensions
          output = output.squeeze(0).squeeze(0);
        
        return output;

    @staticmethod
    def backward(ctx, grad_output):
        r""" Compute df/dx from df/dy using the Chain Rule df/dx = df/dx*dy/dx.
        For periodic padding we use df/dx = sum_{y_i ~ x} df/dy_i, where 
        the y_i ~ x are all points y_i that are equivalent to x under the periodic extension. 
        
        For extracting from periodic tiling the unit cell, the derivatives dy/dx are zero 
        unless x is within the unit cell.  The block matrix is dy/dx = [[Z,Z,Z],[Z,W,Z],[Z,Z,Z]],
        where W is the derivative values in the unit cell (df/dy) and Z is the zero matrix.
        """        
        num_tiles = ctx.num_tiles;
        
        nd = grad_output.dim();
        if (nd > 4):
          raise Exception('Expects tensor that has number of dimensions <= 4 (dim = {}).'.format(nd));
        
        if (nd < 2):
          raise Exception('Expects tensor that has at least number of dimensions >= 2 (dim = {}).'.format(nd));                            

        W = grad_output; # short-hand for grad_output (size of output)
        if (nd == 2):
          # add extra dimensions so we can process all tensors the same way
          W = W.unsqueeze(0).unsqueeze(0); # size --> [1,1,nx,ny]          
                        
        s = W.size();
        Z = grad_output.new_zeros(s[0],s[1],s[2],s[3]);
        A1 = torch.cat([Z,Z,Z],dim=3);
        B1 = torch.cat([Z,W,Z],dim=3);
        bb = torch.cat([A1,B1,A1],dim=2);
        
        grad_input = bb; 
                
        if (nd == 2): # if only [nx,ny] then squeeze out extra dimensions
          grad_input = grad_input.squeeze(0).squeeze(0);
        
        return grad_input;

#******************************************
# Custom Modules
#******************************************
class PeriodicPad2d(nn.Module):
    def __init__(self,flag_coords=False):
        r"""Setup for computing the periodic tiling."""
        super(PeriodicPad2d, self).__init__()
        self.flag_coords = flag_coords;

    def forward(self, input):
        r"""Compute the periodic padding of the input. """
        return PeriodicPad2dFunc.apply(input,None,self.flag_coords);

    def extra_repr(self):
        r"""Displays some of the information associated with the module. """
        return 'PeriodicPad2d: (no internal parameters)';

class ExtractUnitCell2d(nn.Module):
    def __init__(self):
        super(ExtractUnitCell2d, self).__init__()

    def forward(self, input):
        r"""Computes the periodic padding of the input."""
        return ExtractUnitCell2dFunc.apply(input);

    def extra_repr(self):
        r"""Displays some of the information associated with the module. """
        return 'ExtractUnitCell2d: (no internal parameters)';




"""
  Collection of routines helpful for visualizing results and generating figures.
"""

# Authors: B.J. Gross and P.J. Atzberger
# Website: http://atzberger.org/

import matplotlib;
import matplotlib as mtpl;
import matplotlib.pyplot as plt;
import matplotlib.gridspec as gridspec;

import numpy as np;

#----------------------------
def plot_samples_u_f_1d(u_list,f_list,np_xj,np_xi,rows=4,cols=6,
                       figsize = (20*0.9,10*0.9),title="Data Samples: u, f=L[u]",
                       xlabel='x',ylabel='',
                       left=0.125,bottom=0.1,right=0.9, top=0.94,wspace=0.4,hspace=0.4,
                       fontsize=16,y=1.00,flag_draw=True,**extra_params):      

  r"""Plots a collection of data samples in a panel."""
  # --  
  fig, axs = plt.subplots(nrows=rows, ncols=cols, figsize=figsize,sharey=False);
  fig.subplots_adjust(left=left,bottom=bottom,right=right,top=top,wspace=wspace,hspace=hspace);
  plt.suptitle(title, fontsize=fontsize,y=y); 

  # -- 
  I1 = 0; I2 = 0;
  for i1 in range(0,rows):
    for i2 in range(0,cols):          
      ax = axs[i1,i2];  

      if i2 % 2 == 0:
        xx = np_xj[:,0]; yy = u_list[I1].numpy()[0,:];
        #yy = yy.squeeze(0);
        ax.plot(xx,yy,'m.-');
        if i1 == 0:
          ax.set_title('u');
        I1 += 1;
      else:
        xx = np_xi[:,0]; yy = f_list[I2].numpy()[0,:];
        #yy = yy.squeeze(0);
        ax.plot(xx,yy,'r.-');
        if i1 == 0:
          ax.set_title('f');
        I2 += 1;
      
      ax.set_xlabel(xlabel);
      ax.set_ylabel(ylabel);

  if flag_draw:
    plt.draw();

def plot_samples_u_f_fp_1d(u_list,f_target_list,f_pred_list,np_xj,np_xi,rows=4,cols=6,
                           figsize = (20*0.9,10*0.9),title="Data Samples: u, f=L[u]",
                           xlabel='x',ylabel='',
                           left=0.125,bottom=0.1,right=0.9, top=0.94,wspace=0.4,hspace=0.4,
                           fontsize=16,y=1.00,flag_draw=True,**extra_params):      

  r"""Plots a collection of data samples and predictions in a panel."""
  # --  
  fig, axs = plt.subplots(nrows=rows, ncols=cols, figsize=figsize,sharey=False);
  fig.subplots_adjust(left=left,bottom=bottom,right=right,top=top,wspace=wspace,hspace=hspace);
  plt.suptitle(title, fontsize=fontsize,y=y); 

  # -- 
  I1 = 0; I2 = 0;
  for i1 in range(0,rows):
    for i2 in range(0,cols):          
      ax = axs[i1,i2];  
    
      if i2 % 2 == 0:
        ax.plot(np_xj[:,0],u_list[I1].numpy()[0,:],'m.-');
        if i1 == 0:
          ax.set_title('u');
        I1 += 1;  
        ax.set_xlabel(xlabel);
        ax.set_ylabel(ylabel);
      else:                
        ax.plot(np_xj[:,0],f_target_list[I2].numpy()[0,:],'r.-');
        ax.plot(np_xj[:,0],f_pred_list[I2].numpy()[0,:],'b.-');          
        if i1 == 0:
          ax.set_title('f');
        I2 += 1;  
        ax.set_xlabel(xlabel);
        ax.set_ylabel(ylabel);
      
  if flag_draw:
    plt.draw();

def plot_dataset_diffOp1(dataset,np_xj=None,np_xi=None,rows=4,cols=6,II=None,
                         figsize=(20*0.9,10*0.9),title="Data Samples: u, f=L[u]",
                         xlabel='x',ylabel='',
                         left=0.125,bottom=0.1,right=0.9, top=0.94,wspace=0.4,hspace=0.4,
                         fontsize=16,y=1.00,flag_draw=True,**extra_params):

  r"""Plots a collection of data samples in a panel."""    
  u_list = []; f_list = []; f_pred_list = []; 
  num_samples = len(dataset);
  for I in np.arange(0,min(num_samples,int(rows*cols/2))):                
    if II is None:
      u_list.append(dataset[I][0].cpu()); # just make plain lists for convenience
      f_list.append(dataset[I][1].cpu()); 
    else:
      u_list.append(dataset[II[I]][0].cpu());
      f_list.append(dataset[II[I]][1].cpu());
      
  # --
  plot_samples_u_f_1d(u_list=u_list,f_list=f_list,np_xj=np_xj,np_xi=np_xi,
                      rows=rows,cols=cols,
                      figsize=figsize,title=title,
                      xlabel=xlabel,ylabel=ylabel,
                      left=left,bottom=bottom,right=right,top=top,wspace=wspace,hspace=hspace,
                      fontsize=fontsize,y=y,flag_draw=flag_draw,**extra_params);

#----------------------------
def plot_samples_u_f_2d(u_list,f_list,np_xj,np_xi,channelI_u=0,channelI_f=0,rows=4,cols=6,
                       figsize = (20*0.9,10*0.9),title="Data Samples: u, f=L[u]",
                       xlabel='x',ylabel='',
                       left=0.125, bottom=0.1, right=0.9, top=0.94, wspace=0.01, hspace=0.1,
                       fontsize=16,y=1.00,flag_draw=True,**extra_params):      

  r"""Plots a collection of data samples in a panel."""
  # --  
  fig, axs = plt.subplots(nrows=rows, ncols=cols, figsize=figsize,sharey=False);
  fig.subplots_adjust(left=left,bottom=bottom,right=right,top=top,wspace=wspace,hspace=hspace);
  plt.suptitle(title, fontsize=fontsize,y=y); 

  # --
  for ax in axs.flatten():
    ax.axis('off');

  # -- 
  I1 = 0; I2 = 0; Ic_u = channelI_u;Ic_f = channelI_f;
  for i1 in range(0,rows):
    for i2 in range(0,cols):          
      ax = axs[i1,i2];  

      if i2 % 2 == 0:
        uu = u_list[I1][Ic_u,:,:];
        ax.imshow(uu,cmap='Blues_r');
        if i1 == 0:
          ax.set_title('u');
        I1 += 1;  
      else:                
        ff = f_list[I2][Ic_f,:,:];
        ax.imshow(ff,cmap='Purples_r');
        if i1 == 0:
          ax.set_title('f');
        I2 += 1;   

  if flag_draw:
    plt.draw();

def plot_samples_u_f_fp_2d(u_list,f_target_list,f_pred_list,np_xj,np_xi,channelI_u=0,channelI_f=0,rows=4,cols=6,
                           figsize = (20*0.9,10*0.9),title="Data Samples: u, f=L[u]",
                           xlabel='x',ylabel='',
                           left=0.125, bottom=0.1, right=0.9, top=0.94, wspace=0.01, hspace=0.1,
                           fontsize=16,y=1.00,flag_draw=True,**extra_params):      

  r"""Plots a collection of data samples and predictions in a panel."""
  # --  
  fig, axs = plt.subplots(nrows=rows, ncols=cols, figsize=figsize,sharey=False);
  fig.subplots_adjust(left=left,bottom=bottom,right=right,top=top,wspace=wspace,hspace=hspace);
  plt.suptitle(title, fontsize=fontsize,y=y); 

  # --
  for ax in axs.flatten():
    ax.axis('off');

  # -- 
  Ic_u = channelI_u; Ic_f = channelI_f;
  I1 = 0; I2 = 0; I3 = 0; 
  for i1 in range(0,rows):
    for i2 in range(0,cols):          
      ax = axs[i1,i2];  

      if i2 % 3 == 0:
        uu = u_list[I1][Ic_u,:,:];
        ax.imshow(uu,cmap='Blues_r');
        if i1 == 0:
          ax.set_title('u');
        I1 += 1;  
      elif i2 % 3 == 1: 
        ff = f_pred_list[I2][Ic_f,:,:];
        ax.imshow(ff,cmap='Purples_r');
        if i1 == 0:
          ax.set_title('f:predicted');
        I2 += 1;  
      elif i2 % 3 == 2: 
        ff = f_target_list[I3][Ic_f,:,:];
        ax.imshow(ff,cmap='Purples_r');
        if i1 == 0:
          ax.set_title('f:target');
        I3 += 1;              
      
  if flag_draw:
    plt.draw();

def plot_dataset_diffOp2(dataset,np_xj=None,np_xi=None,channelI_u=0,channelI_f=0,rows=4,cols=6,II=None,
                         figsize=(20*0.9,10*0.9),title="Data Samples: u, f=L[u]",
                         xlabel='x',ylabel='',
                         left=0.125, bottom=0.1, right=0.9, top=0.94, wspace=0.01, hspace=0.1,
                         fontsize=16,y=1.00,flag_draw=True,**extra_params):



  r"""Plots a collection of data samples in a panel."""    
  u_list = []; f_list = []; f_pred_list = []; 
  num_samples = len(dataset);
  for I in np.arange(0,min(num_samples,int(rows*cols/2))):                
    if II is None:
      u_list.append(dataset[I][0].cpu()); # just make plain lists for convenience
      f_list.append(dataset[I][1].cpu()); 
    else:
      u_list.append(dataset[II[I]][0].cpu());
      f_list.append(dataset[II[I]][1].cpu());
      
  # --
  plot_samples_u_f_2d(u_list=u_list,f_list=f_list,np_xj=np_xj,np_xi=np_xi,
                      channelI_u=channelI_u,channelI_f=channelI_f,
                      rows=rows,cols=cols,
                      figsize=figsize,title=title,
                      xlabel=xlabel,ylabel=ylabel,
                      left=left,bottom=bottom,right=right,top=top,wspace=wspace,hspace=hspace,
                      fontsize=fontsize,y=y,flag_draw=flag_draw,**extra_params);

#----------------------------
def plot_images_in_array(axs,img_arr,label_arr=None,cmap=None, **extra_params):
    r"""Plots an array of images as a collection of panels."""
    
    numSamples = len(img_arr);
    sqrtS      = int(np.sqrt(numSamples));
        
    # Default values
    flag_plot_rect = False;
    list_correct_class = None;
    
    if 'list_correct_class' in extra_params:
      list_correct_class = extra_params['list_correct_class'];
      flag_plot_rect     = True;
            
    if 'flag_plot_rect' in extra_params:
      flag_plot_rect = extra_params['flag_plot_rect'];
               
    I = 0;
    for i in range(0,sqrtS):        
        for j in range(0,sqrtS):
            ax = axs[i][j];
            img = img_arr[I];            
            
            if len(img.shape) >= 3: 
              if img.shape[2] == 1: # For BW case of (Nx,Ny,1) --> (Nx,Ny), RGB has (Nx,Ny,3).
                img = img.squeeze(2);
                                        
            if cmap is not None:
              ax.imshow(img, cmap=cmap);
            else:
              ax.imshow(img);
            
            if label_arr is not None:
              ax.set_title("%s"%label_arr[I]);
            
            ax.set_xticks([]);
            ax.set_yticks([]);
                        
            if flag_plot_rect:
                
              if list_correct_class[I]:
                edge_color = 'g';
              else:
                edge_color = 'r';
                
              # draw a rectangle 
              Nx = img.shape[0]; Ny = img.shape[1];    
              rectangle = mtpl.patches.Rectangle((0,0),Nx-1,Ny-1,
                                                 linewidth=5,edgecolor=edge_color,
                                                 facecolor='none');              
              ax.add_patch(rectangle);
            
            I += 1;

def plot_image_array(img_arr,label_arr=None,title=None,figSize=(18,18),
                     title_yp=0.95,cmap="gray",**extra_params):
    r"""Plots an array of images as a collection of panels."""

    # determine number of images we need to plot
    numSamples = len(img_arr);
    sqrtS      = int(np.sqrt(numSamples));
    rows       = sqrtS;
    cols       = sqrtS;
    
    fig, axs = plt.subplots(nrows=rows, ncols=cols, figsize=figSize);
        
    plot_images_in_array(axs,img_arr,label_arr,cmap=cmap,**extra_params);

    if title is None:
      plt.suptitle("Collection of Images", fontsize=18,y=title_yp);
    else:
      plt.suptitle(title, fontsize=18,y=title_yp);
    
def plot_gridspec_image_array(outer_h,img_arr,label_arr=None,title=None,figSize=(18,18),title_yp=0.95,cmap="gray",title_x=0.0,title_y=1.0,title_fsize=14):    
    r"""Plots an array of images as a collection of panels."""  
    fig = plt.gcf();
    
    sqrtS = int(np.sqrt(len(img_arr)));

    inner = gridspec.GridSpecFromSubplotSpec(sqrtS, sqrtS,
                                             subplot_spec=outer_h, wspace=0.1, hspace=0.1);
    
    # collect the axes
    axs =[];
    I = 0;
    for i in range(0,sqrtS):
      axs_r = [];
      for j in range(0,sqrtS):
        ax = plt.Subplot(fig, inner[I]);
        fig.add_subplot(ax);
        axs_r.append(ax);
        I += 1;    
        
      axs.append(axs_r);

    # plot the images
    plot_images_in_array(axs,img_arr,cmap="gray");
    
    if title is None:      
      a = 1;
    else:
      axs[0][0].text(title_x,title_y,title,fontsize=title_fsize);

#----------------------------
def save_fig(baseFilename,extraLabel,flag_verbose=0,dpi_set=200,flag_pdf=False):
  r"""Saves figures to disk."""
    
  fig = plt.gcf();  
  fig.patch.set_alpha(1.0);
  fig.patch.set_facecolor((1.0,1.0,1.0,1.0));
  
  if flag_pdf:
    saveFilename = '%s%s.pdf'%(baseFilename,extraLabel);
    if flag_verbose > 0:
      print('saveFilename = %s'%saveFilename);
    plt.savefig(saveFilename, format='pdf',dpi=dpi_set,facecolor=(1,1,1,1),alpha=1.0);

  saveFilename = '%s%s.png'%(baseFilename,extraLabel);
  if flag_verbose > 0:
    print('saveFilename = %s'%saveFilename);
  plt.savefig(saveFilename, format='png',dpi=dpi_set,facecolor=(1,1,1,1),alpha=1.0);





import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gmlsnets_pytorch",
    version="1.0.0",
    author="Ben J. Gross and Paul J. Atzberger",
    author_email="atzberg@gmail.com",
    description="Implementation of GMLS-Nets in PyTorch.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://atzberger.org/",
    packages=setuptools.find_packages(include=['gmlsnets_pytorch']),
    #setup_requires=["numpy>=1.16","scipy>=1.3","matplotlib>=3.0"],
    #install_requires=["numpy>=1.16","scipy>=1.3","matplotlib>=3.0"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)

